import os
import re
from fsbc.application import app
from fsbc.settings import Settings
from fsgs.BaseDatabase import BaseDatabase
from fsgs.FSGSDirectories import FSGSDirectories
import threading

thread_local = threading.local()
VERSION = 40
RESET_VERSION = 39
QUOTED_TERMS_RE = re.compile('["].*?["]')


class Database(BaseDatabase):

    VERSION = VERSION
    RESET_VERSION = RESET_VERSION
    GAME_LIST_GAMES = "cbc209ef-c93d-4db7-be52-c159bfec43dc"
    GAME_LIST_CONFIGS = "106409c1-dc49-4601-8e47-8cf6780ddb3b"

    @classmethod
    def get_path(cls):
        return os.path.join(FSGSDirectories.databases_dir(), "Launcher.sqlite")

    def __init__(self, sentinel):
        BaseDatabase.__init__(self, sentinel)

    def get_version(self):
        return VERSION

    def get_reset_version(self):
        return RESET_VERSION

    @classmethod
    def instance(cls, new=False):
        if new or not hasattr(cls.thread_local, "database"):
            cls.thread_local.database = cls(cls.SENTINEL)
        return cls.thread_local.database

    @classmethod
    def get_instance(cls):
        if not hasattr(cls.thread_local, "database"):
            cls.thread_local.database = cls(cls.SENTINEL)
        return cls.thread_local.database

    def __del__(self):
        print("Database.__del__")

    # def init(self):
    #     if self.connection:
    #         return
    #     self.connection = sqlite3.connect(self.get_database_path())
    #     self._cursor = self.connection.cursor()
    #     self._cursor = self._cursor
    #     self.ensure_updated_database()
    #
    # def cursor(self):
    #     self.init()
    #     return self.connection.cursor()

    def get_files(self, ext=None):
        self.init()
        query = "SELECT path, name FROM file WHERE 1 = 1 "
        args = []
        if ext is not None:
            query += " AND path like ?"
            args.append("%" + ext)
        cursor = self.internal_cursor()
        cursor.execute(query, args)
        results = []
        for row in cursor:
            data = {"path": self.decode_path(row[0]), "name": row[1]}
            results.append(data)
        return results

    def get_configuration_path(self, id):
        self.init()
        query = "SELECT path FROM configuration WHERE id = ?"
        cursor = self.internal_cursor()
        cursor.execute(query, (id,))
        path = self.decode_path(cursor.fetchone()[0])
        return path

    # def get_config(self, id):
    #     self.init()
    #     query = "SELECT name, uuid, path, data, game_rating.work_rating, " \
    #             "game_rating.like_rating FROM configuration LEFT JOIN " \
    #             "game_rating ON game_rating.game = uuid WHERE id = ?"
    #     self._cursor.execute(query, (id,))
    #     row = self._cursor.fetchone()
    #     return {
    #         "name": row[0],
    #         "uuid": row[1],
    #         "path": self.decode_path(row[2]),
    #         "data": row[3],
    #         "work_rating": row[4],
    #         "like_rating": row[5],
    #     }

    # def get_game_info(self, id):
    #     self.init()
    #     query = "SELECT name, uuid, path FROM game WHERE id = ?"
    #     self._cursor.execute(query, (id,))
    #     row = self._cursor.fetchone()
    #     return {
    #         "name": row[0],
    #         "uuid": row[1],
    #         "path": self.decode_path(row[2]),
    #     }

    def encode_path(self, path):
        # this only works if both path and FSGSDirectories.base_dir
        # (etc) have been normalized with get_real_case.
        path = path.replace("\\", "/")
        base_dir = FSGSDirectories.get_base_dir()
        if path.startswith(base_dir):
            path = path[len(base_dir) :]
            if path.startswith("/"):
                path = path[1:]
            path = "$/" + path
        return path

    def decode_path(self, path):
        if not path or path[0] != "$":
            return path
        base_dir = FSGSDirectories.get_base_dir() + "/"
        if path.startswith("$/"):
            path = base_dir + path[2:]
        return path

    def find_local_configurations(self):
        cursor = self.internal_cursor()
        # query = "SELECT id, path FROM configuration WHERE path like ?"
        # args = ["$BASE/Configurations/%"]
        # self._cursor.execute(query, args)
        a = "$/Configurations/"
        b = "$/Configurations" + "\u0030"  # one more than forward slash
        query = "SELECT id, path FROM game WHERE " "path >= ? AND path < ?"
        cursor.execute(query, (a, b))
        result = {}
        for row in cursor.fetchall():
            result[self.decode_path(row[1])] = row[0]
        return result

    # def delete_configuration(self, id=-1, path=None):
    #     self.init()
    #     if path is not None:
    #         query = "DELETE FROM configuration WHERE path = ?"
    #         path = self.encode_path(path)
    #         args = [path]
    #     else:
    #         query = "DELETE FROM configuration WHERE id = ?"
    #         args = [id]
    #     self._cursor.execute(query, args)

    # def find_local_roms(self):
    #     self.init()
    #
    #     a = "$BASE/Kickstarts/"
    #     b = "$BASE/Kickstarts" + "\u0030" # one more than forward slash
    #     query = "SELECT id, path FROM file WHERE path >= ? AND path < ?"
    #     #args = ["$BASE/Kickstarts/%"]
    #     self._cursor.execute(query, (a, b))
    #     result = {}
    #     for row in self._cursor.fetchall():
    #         result[self.decode_path(row[1])] = row[0]
    #     return result
    #
    # def delete_file(self, id=-1, path=None):
    #     self.init()
    #     if path is not None:
    #         query = "DELETE FROM file WHERE path = ?"
    #         path = self.encode_path(path)
    #         args = [path]
    #     else:
    #         query = "DELETE FROM file WHERE id = ?"
    #         args = [id]
    #     self._cursor.execute(query, args)

    # def search_configurations(self, search):
    #     self.init()
    #     query = "SELECT id, name, type, sort_key FROM configuration WHERE " \
    #             "type < 5"
    #     args = []
    #     for word in search.split(" "):
    #         word = word.strip().lower()
    #         if word:
    #             #if len(args) == 0:
    #             #    query = query + " WHERE search like ?"
    #             #else:
    #             query = query + " AND search like ?"
    #             args.append("%{0}%".format(word))
    #     query = query + " ORDER BY sort_key"
    #     cursor = self.internal_cursor()
    #     cursor.execute(query, args)
    #     return cursor.fetchall()

    # def find_game_variants(self, game_uuid):
    #     self.init()
    #     query = "SELECT id, name, uuid, configuration.like_rating, " \
    #             "configuration.work_rating, game_rating.like_rating, " \
    #             "have FROM configuration LEFT JOIN " \
    #             "game_rating ON game_rating.game = uuid WHERE " \
    #             "reference = ? AND have >= ?"
    #     query += " ORDER BY name"
    #     print(query, game_uuid)
    #     cursor = self.internal_cursor()
    #     cursor.execute(query, (game_uuid,))
    #     return cursor.fetchall()

    def find_game_database_for_game_variant(self, uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT database FROM game_variant WHERE " "uuid = ?", (uuid,)
        )
        row = cursor.fetchone()
        if row is None:
            raise LookupError("game variant not found")
        return row[0]

    def find_game_variants_new(self, game_uuid="", have=3):
        include_unpublished = False
        if Settings.instance()["database_show_unpublished"] == "1":
            include_unpublished = True
        cursor = self.internal_cursor()
        print("FIXME: not looking up ratings yet!")
        query = (
            "SELECT uuid, name, game_uuid, 0 as like_rating, "
            "0 as work_rating, have, database, published "
            "FROM game_variant WHERE "
            "game_uuid = ? AND have >= ?"
        )
        if not include_unpublished:
            query += " AND (published = 1 OR published IS NULL)"
        query += " ORDER BY like_rating DESC, work_rating DESC, name"
        cursor.execute(query, (game_uuid, have))
        result = []
        for row in cursor:
            result.append(dict(row))
        return result

    def get_last_game_variant(self, game_uuid):
        query = "SELECT variant_uuid FROM last_variant WHERE game_uuid = ?"
        cursor = self.internal_cursor()
        cursor.execute(query, (game_uuid,))
        row = cursor.fetchone()
        if row:
            return row[0]
        return ""

    def set_last_game_variant(self, game_uuid, variant_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "DELETE FROM last_variant WHERE game_uuid = ?", (game_uuid,)
        )
        cursor.execute(
            "INSERT INTO last_variant (game_uuid, variant_uuid) "
            "VALUES (?, ?)",
            (game_uuid, variant_uuid),
        )

    def search_games(self, search):
        self.init()
        query = "SELECT id, name FROM game"
        args = []
        for word in search.split(" "):
            word = word.strip().lower()
            if word:
                if len(args) == 0:
                    query += " WHERE search like ?"
                else:
                    query += " AND search like ?"
                args.append("%{0}%".format(word))
        query += " ORDER BY name"
        cursor = self.internal_cursor()
        cursor.execute(query, args)
        return cursor.fetchall()

    def find_game(self, uuid="", result=None):
        self.init()
        cursor = self.internal_cursor()
        if uuid:
            cursor.execute(
                "SELECT path FROM game WHERE uuid = ? LIMIT 1", (uuid,)
            )
        row = cursor.fetchone()
        if row:
            path = self.decode_path(row[0])
            if result is not None:
                result["path"] = path
            return path
        else:
            if result is not None:
                result["path"] = None
            return None

    def find_file(self, name="", sha1="", path="", result=None):
        cursor = self.internal_cursor()
        if sha1:
            # print("xxx", repr(sha1))
            # import traceback
            # traceback.print_stack()
            # print("check sha1")
            cursor.execute(
                "SELECT id, path, sha1, mtime, size FROM file "
                "WHERE sha1 = ? LIMIT 1",
                (sha1,),
            )
        elif name:
            # print("check name")
            cursor.execute(
                "SELECT id, path, sha1, mtime, size FROM file "
                "WHERE name = ? COLLATE NOCASE LIMIT 1",
                (name.lower(),),
            )
        else:
            path = self.encode_path(path)
            cursor.execute(
                "SELECT id, path, sha1, mtime, size FROM file "
                "WHERE path = ? LIMIT 1",
                (path,),
            )
        row = cursor.fetchone()
        # print("---------", row)
        if row:
            path = self.decode_path(row[1])
            if result is not None:
                result["id"] = row[0]
                result["path"] = path
                result["sha1"] = row[2]
                result["mtime"] = row[3]
                result["size"] = row[4]
            return path
        else:
            if result is not None:
                result["id"] = None
                result["path"] = None
                result["sha1"] = None
                result["mtime"] = None
                result["size"] = None
            return None

    def add_file(
        self,
        path="",
        sha1=None,
        md5=None,
        crc32=None,
        mtime=0,
        size=0,
        scan=0,
        name="",
    ):
        cursor = self.internal_cursor()
        if not name:
            name = os.path.basename(path)
        path = self.encode_path(path)

        # print("adding path", path)
        # p, name = os.path.split(path)
        cursor.execute(
            "INSERT INTO file (path, sha1, mtime, size, "
            "md5, crc32, name, scan) VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
            (path, sha1, mtime, size, md5, crc32, name, scan),
        )

    # def add_configuration(
    #         self, path="", uuid="", data="", name="", search="", scan=0,
    #         type=0, reference=None, like_rating=0, work_rating=0,
    #         sort_key=""):
    #     cursor = self.internal_cursor()
    #     if not sort_key:
    #         sort_key = name.lower()
    #     path = self.encode_path(path)
    #     cursor.execute(
    #         "INSERT INTO configuration (path, name, scan, "
    #         "search, uuid, data, type, reference, like_rating, "
    #         "work_rating, sort_key) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, "
    #         "?, ?)",
    #         (path, name, scan, search, uuid, data, type, reference,
    #          like_rating, work_rating, sort_key))

    # def ensure_game_configuration(self, uuid, name, sort_key, scan=0, type=1):
    #     cursor = self.internal_cursor()
    #     cursor.execute(
    #         "SELECT * FROM configuration WHERE uuid = ? "
    #         "AND name = ? AND sort_key = ? AND scan = ? AND type < ?",
    #         (uuid, name, sort_key, scan, type))
    #     row = cursor.fetchone()
    #     if row is None:
    #         cursor.execute("DELETE from configuration WHERE uuid = ?",
    #             (uuid,))
    #         search = name.lower()
    #         self.add_configuration(
    #             uuid=uuid, name=name, search=search, scan=scan, type=type,
    #             sort_key=sort_key)

    def add_configuration(self, path="", name="", sort_key=""):
        cursor = self.internal_cursor()
        path = self.encode_path(path)
        if not sort_key:
            sort_key = name.lower()
        cursor.execute(
            "SELECT id, have, name, sort_key FROM game WHERE path = ?", (path,)
        )
        row = cursor.fetchone()
        if row:
            game_id = row[0]
            if row[1] == 5 and row[2] == name and row[3] == sort_key:
                # already up to date
                return game_id
        else:
            cursor.execute("INSERT INTO game (path) VALUES (?)", (path,))
            game_id = cursor.lastrowid
        cursor.execute(
            "UPDATE game SET have = 5, name = ?, sort_key = ?, "
            "published = 1, adult = 0 WHERE id = ?",
            (name, sort_key, game_id),
        )
        return game_id

    def delete_game(self, id):
        cursor = self.internal_cursor()
        cursor.execute("DELETE FROM search_term WHERE game = ?", (id,))
        cursor.execute("DELETE FROM game WHERE id = ?", (id,))

    def update_game_search_terms(self, game_id, search_terms):
        cursor = self.internal_cursor()
        search_terms = sorted(search_terms)
        cursor.execute(
            "SELECT term FROM search_term WHERE game = ?", (game_id,)
        )
        existing_terms = sorted(x[0] for x in cursor)
        if search_terms != existing_terms:
            cursor.execute(
                "DELETE FROM search_term WHERE game = ?", (game_id,)
            )
            for term in search_terms:
                cursor.execute(
                    "INSERT INTO search_term (game, " "term) VALUES (?, ?)",
                    (game_id, term),
                )

    # def remove_unscanned_configurations(self, scan):
    #     cursor = self.internal_cursor()
    #     cursor.execute("DELETE FROM configuration WHERE scan != ?", (scan,))

    # def remove_unscanned_games(self, scan):
    #     self.init()
    #     self._cursor.execute("DELETE FROM game WHERE scan != ?",
    #             (scan,))

    def get_variant_for_list_and_game(self, list_uuid, game_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT variant_uuid FROM game_list_game WHERE "
            "list_uuid = ? AND game_uuid = ?",
            (list_uuid, game_uuid),
        )
        row = cursor.fetchone()
        if row is None:
            return None
        return row[0]

    def find_games_new(
        self, search="", have=3, list_uuid="", database_only=False
    ):
        print("Database.find_games_new search = {0}".format(repr(search)))
        non_database_only = False
        if list_uuid == self.GAME_LIST_GAMES:
            database_only = True
            list_uuid = ""
        elif list_uuid == self.GAME_LIST_CONFIGS:
            non_database_only = True
            list_uuid = ""
        elif list_uuid:
            have = 0

        cursor = self.internal_cursor()
        query = (
            "SELECT DISTINCT uuid, name, platform, year, publisher, "
            "front_image, title_image, screen1_image, screen2_image, "
            "screen3_image, screen4_image, screen5_image, have, path, "
            "sort_key, subtitle, thumb_image, backdrop_image, "
            "published FROM game"
        )

        args = []
        if list_uuid:
            query += (
                " INNER JOIN game_list_game "
                "ON game.uuid = game_list_game.game_uuid "
            )

        have_false = False
        published_false = False
        additional_clauses = []
        additional_args = []
        include_adult = False
        if app.settings["database_show_adult"] == "1":
            include_adult = True
        include_unpublished = False
        if app.settings["database_show_unpublished"] == "1":
            include_unpublished = True

        terms = []

        def callback(m):
            terms.append(m.group(0))
            return ""

        stripped_search = QUOTED_TERMS_RE.sub(callback, search)
        for term in stripped_search.split(" "):
            terms.append(term)

        for i, term in enumerate(terms):
            term = term.strip().lower()
            exact_term = False
            if term.startswith('"'):
                exact_term = True
                term = term[1:]
            if term.endswith('"'):
                term = term[:-1]

            if term.startswith("platform:"):
                term = term.replace("platform:", "s:")
                exact_term = True
            elif term.startswith("players:"):
                term = term.replace("players:", "p:")
                exact_term = True
            elif term.startswith("year:"):
                term = term.replace("year:", "y:")
                exact_term = True
            elif term.startswith("tag:"):
                term = term.replace("tag:", "t:")
                exact_term = True
                if term == "t:adult":
                    include_adult = True
                # if term == "t:unpublished":
                #     include_unpublished = True
            elif term.startswith("letter:"):
                exact_term = True
                term = term.replace("letter:", "l:")
            elif term == "have:false":
                have_false = True
                continue
            elif term == "published:false":
                published_false = True
                continue

            if term.startswith("s:"):
                from fsgs.platform import normalize_platform_id

                term = "s:" + normalize_platform_id(term[2:])

            if term:
                # searching for quoted terms with space won't generally work
                # if " " in term:
                #     #additional_clauses.append(" AND search like ?")
                #     #additional_args.append("%" + term + "%")
                if exact_term:
                    query += (
                        " INNER JOIN search_term as st{0} ON "
                        "game.id = st{0}.game AND "
                        "st{0}.term = ?".format(i)
                    )
                    args.append(term)
                else:
                    query += (
                        " INNER JOIN search_term as st{0} ON "
                        "game.id = st{0}.game AND "
                        "st{0}.term like ?".format(i)
                    )
                    args.append(term + "%")

        if have_false:
            query += " WHERE have = 0"
        else:
            query += " WHERE have >= {0}".format(int(have))
        if published_false:
            query += " AND published = 0"
        elif not include_unpublished:
            query += " AND published = 1"
        if not include_adult:
            query += " AND adult = 0"
        for clause in additional_clauses:
            query += clause
        if list_uuid:
            query += " AND game_list_game.list_uuid = ?"
            args.append(list_uuid)
        if database_only:
            query += " AND path is NULL"
        elif non_database_only:
            query += " AND path is NOT NULL"
        query += " ORDER BY"
        if list_uuid:
            query += " game_list_game.position,"
        args.extend(additional_args)
        query += " sort_key, platform"

        print(query.replace("?", "{}").format(*args))
        cursor.execute(query, args)
        return cursor.fetchall()

    # def add_game_new(self, uuid, name, platform, year, publisher, front_image,
    #                  title_image, screen1_image, screen2_image, screen3_image,
    #                  screen4_image, screen5_image, sort_key, scanned=0):
    #     cursor = self.internal_cursor()
    #     cursor.execute(
    #         "SELECT * FROM game WHERE uuid = ? "
    #         "AND name = ? AND platform = ? AND year = ? "
    #         "AND publisher = ? AND front_image = ? AND title_image = ? "
    #         "AND screen1_image = ? AND screen2_image = ? "
    #         "AND screen3_image = ? AND screen4_image = ? "
    #         "AND screen5_image = ? AND sort_key = ? AND scanned = ?",
    #         (uuid, name, platform, year, publisher, front_image,
    #         title_image, screen1_image, screen2_image, screen3_image,
    #         screen4_image, screen5_image, sort_key, scanned))
    #     row = cursor.fetchone()
    #     if row is None:
    #         cursor.execute("DELETE from game WHERE uuid = ?", (uuid,))
    #         cursor.execute(
    #             "INSERT INTO game(uuid, name, "
    #             "platform, year, publisher, front_image, title_image, "
    #             "screen1_image, screen2_image, screen3_image, "
    #             "screen4_image, screen5_image, sort_key, scanned) "
    #             "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
    #             (uuid, name, platform, year, publisher, front_image,
    #             title_image, screen1_image, screen2_image, screen3_image,
    #             screen4_image, screen5_image, sort_key, scanned))

    # def add_game_variant_new(
    #         self, uuid="", name="", game_uuid="", like_rating=0,
    #         work_rating=0, scanned=0):
    #     cursor = self.internal_cursor()
    #     query = "INSERT INTO game_variant(uuid, name, game_uuid, " \
    #             "like_rating, work_rating, scanned) VALUES " \
    #             "(?, ?, ?, ?, ?, ?)"
    #     args = (uuid, name, game_uuid, like_rating, work_rating, scanned)
    #     cursor.execute(query, args)
    #     return cursor.fetchall()

    def get_ratings_for_game(self, game_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT like_rating, work_rating FROM rating WHERE "
            "game_uuid = ?",
            (game_uuid,),
        )
        row = cursor.fetchone()
        if row is not None:
            return row[0], row[1]
        else:
            return 0, 0

    def get_last_file_event_stamps(self):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT last_file_insert, last_file_delete, "
            "database_locker FROM metadata"
        )
        row = cursor.fetchone()
        result = {
            "last_file_insert": row[0],
            "last_file_delete": row[1],
            "database_locker": row[2],
        }
        return result

    def update_last_file_event_stamps(self, stamps):
        cursor = self.internal_cursor()
        last_stamps = self.get_last_file_event_stamps()
        if stamps["last_file_insert"] != last_stamps["last_file_insert"]:
            cursor.execute(
                "UPDATE metadata set last_file_insert = ?",
                (stamps["last_file_insert"],),
            )
        if stamps["last_file_delete"] != last_stamps["last_file_delete"]:
            cursor.execute(
                "UPDATE metadata set last_file_delete = ?",
                (stamps["last_file_delete"],),
            )
        if stamps["database_locker"] != last_stamps["database_locker"]:
            cursor.execute(
                "UPDATE metadata set database_locker = ?",
                (stamps["database_locker"],),
            )

    def get_game_lists(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT uuid, name FROM game_list")
        return cursor.fetchall()

    def update_database_to_version_39(self):
        cursor = self.internal_cursor()
        cursor.execute(
            """CREATE TABLE game (
                id INTEGER PRIMARY KEY,
                uuid TEXT,
                name TEXT,
                subtitle TEXT,
                platform TEXT,
                year INTEGER,
                publisher TEXT,
                front_image TEXT,
                title_image TEXT,
                screen1_image TEXT,
                screen2_image TEXT,
                screen3_image TEXT,
                screen4_image TEXT,
                screen5_image TEXT,
                thumb_image TEXT,
                backdrop_image TEXT,
                sort_key TEXT,
                have INTEGER,
                path TEXT,
                adult INTEGER,
                published INTEGER,
                update_stamp INTEGER
        )"""
        )
        cursor.execute("CREATE INDEX game_uuid ON game(uuid)")
        cursor.execute("CREATE INDEX game_sort_key ON game(sort_key)")
        cursor.execute(
            """CREATE TABLE game_variant (
                id INTEGER PRIMARY KEY,
                database TEXT,
                uuid TEXT,
                name TEXT,
                game_uuid TEXT,
                work_rating INTEGER,
                like_rating INTEGER,
                have INTEGER,
                published INTEGER,
                update_stamp INTEGER
        )"""
        )
        cursor.execute("CREATE INDEX game_variant_uuid ON game_variant(uuid)")
        cursor.execute(
            "CREATE INDEX game_variant_game_uuid " "ON game_variant(game_uuid)"
        )
        cursor.execute("CREATE INDEX game_path ON game(path)")
        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN last_file_insert INTEGER"
        )
        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN last_file_delete INTEGER"
        )
        cursor.execute(
            """CREATE TABLE search_term (
                game INTEGER,
                term TEXT
        )"""
        )
        cursor.execute(
            """
            CREATE TABLE rating (
                game_uuid VARCHAR(36) PRIMARY KEY NOT NULL,
                work_rating INT NOT NULL,
                like_rating INT NOT NULL,
                updated TIMESTAMP NULL
            );
        """
        )
        cursor.execute(
            """CREATE TABLE last_variant (
            game_uuid CHAR(36) PRIMARY KEY,
            variant_uuid CHAR(36) NOT NULL
            )"""
        )
        cursor.execute(
            """CREATE TABLE game_list (
            uuid CHAR(36) PRIMARY KEY,
            name TEXT,
            sync VARCHAR(36)
            )"""
        )
        cursor.execute(
            """CREATE TABLE game_list_game (
            list_uuid CHAR(36),
            game_uuid CHAR(36),
            variant_uuid CHAR(36),
            position DOUBLE
            )"""
        )
        cursor.execute(
            """CREATE INDEX game_list_game_list_uuid
            ON  game_list_game(list_uuid)"""
        )

    def update_database_to_version_40(self):
        cursor = self.internal_cursor()
        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN database_locker INTEGER"
        )



from fsbc.paths import Paths
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.amiga import Amiga
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.ui.LauncherFilePicker import LauncherFilePicker


class CDManager:
    @classmethod
    def clear_all(cls):
        for i in range(1):
            cls.eject(i)
        cls.clear_cdrom_list()

    @classmethod
    def eject(cls, drive):
        values = [
            ("cdrom_drive_{0}".format(drive), ""),
            ("x_cdrom_drive_{0}_sha1".format(drive), ""),
        ]
        LauncherConfig.set_multiple(values)

    @classmethod
    def clear_cdrom_list(cls):
        values = []
        for i in range(Amiga.MAX_CDROM_IMAGES):
            values.append(("cdrom_image_{0}".format(i), ""))
            values.append(("x_cdrom_image_{0}_sha1".format(i), ""))
        LauncherConfig.set_multiple(values)

    @classmethod
    def multi_select(cls, parent=None):
        default_dir = FSGSDirectories.get_cdroms_dir()
        dialog = LauncherFilePicker(
            parent, gettext("Select Multiple CD-ROMs"), "cd", multiple=True
        )

        if not dialog.show_modal():
            return
        paths = dialog.get_paths()
        paths.sort()

        # checksum_tool = ChecksumTool(parent)
        for i, path in enumerate(paths):
            # sha1 = checksum_tool.checksum(path)
            sha1 = ""
            print("FIXME: not calculating CD checksum just yet")
            path = Paths.contract_path(path, default_dir)

            if i < 1:
                LauncherConfig.set_multiple(
                    [
                        ("cdrom_drive_{0}".format(i), path),
                        ("x_cdrom_drive_{0}_sha1".format(i), sha1),
                    ]
                )
            LauncherConfig.set_multiple(
                [
                    ("cdrom_image_{0}".format(i), path),
                    ("x_cdrom_image_{0}_sha1".format(i), sha1),
                ]
            )

        # blank the rest of the drives
        for i in range(len(paths), 1):
            LauncherConfig.set_multiple(
                [
                    ("cdrom_drive_{0}".format(i), ""),
                    ("x_cdrom_drive_{0}_sha1".format(i), ""),
                ]
            )

            # Config.set("x_cdrom_drive_{0}_sha1".format(i), "")
            # Config.set("x_cdrom_drive_{0}_name".format(i), "")
        # blank the rest of the image list
        for i in range(len(paths), Amiga.MAX_CDROM_IMAGES):
            LauncherConfig.set_multiple(
                [
                    ("cdrom_image_{0}".format(i), ""),
                    ("x_cdrom_image_{0}_sha1".format(i), ""),
                ]
            )



from .client import OGDClient


class SynchronizerContext(object):
    meta = {}

    def __init__(self):
        pass

    @property
    def username(self):
        return OGDClient.auth()[0]

    @property
    def password(self):
        return OGDClient.auth()[1]



import fnmatch
import struct

import KickRom
import RemusFile
import amitools.util.DataDir as DataDir
from amitools.binfmt.BinImage import *


class RomSplitter:
  def __init__(self, split_data_path=None):
    # get data file path
    if split_data_path is None:
      split_data_path = DataDir.ensure_data_sub_dir("splitdata")
    # setup remus file set
    self.rfs = RemusFile.RemusFileSet()
    self.rfs.load(split_data_path)
    # state
    self.chk_sum = None
    self.rom_data = None
    self.remus_rom = None

  def list_roms(self, out, query=None, show_entries=False):
    roms = self.rfs.get_roms()
    for rom in roms:
      if query is None or fnmatch.fnmatch(rom.name, query):
        self.remus_rom = rom
        self.print_rom(out, show_entries)

  def find_rom(self, rom_path):
    """load ROM and try to find a matching dat file.
       Returns True if ROM was matched"""
    self.rom_data = KickRom.Loader.load(rom_path)
    # get check sum
    kh = KickRom.KickRomAccess(self.rom_data)
    if kh.is_kick_rom():
      self.chk_sum = kh.read_check_sum()
    else:
      self.chk_sum = kh.calc_check_sum()
    # search rom in Remus data base
    self.remus_rom = self.rfs.find_rom(self.rom_data, self.chk_sum)
    return self.remus_rom

  def print_rom(self, out, show_entries=False):
    rom = self.remus_rom
    out("@%08x  +%08x  sum=%08x  sum_off=%08x  %s" % \
        (rom.base_addr, rom.size, rom.chk_sum, rom.sum_off, rom.name))
    if show_entries:
      for e in rom.entries:
        self.print_entry(out, e)

  def print_entry(self, out, entry):
    out("  @%06x  +%06x  =%06x  relocs=#%5d  %s" % \
        (entry.offset, entry.size, entry.offset+entry.size,
         len(entry.relocs), entry.name))

  def print_entries(self, out, entries):
    for e in entries:
      self.print_entry(out, e)

  def get_all_entries(self):
    return self.remus_rom.entries

  def query_entries(self, query_str):
    res = []
    for e in self.remus_rom.entries:
      if fnmatch.fnmatch(e.name, query_str):
        res.append(e)
    return res

  def extract_entry(self, entry):
    """return data, relocs"""
    data = self.rom_data[entry.offset:entry.offset+entry.size]
    relocs = entry.relocs
    entry_addr = self.remus_rom.base_addr + entry.offset
    data = self._clean_relocs(data, relocs, entry_addr)
    return data, relocs

  def _clean_relocs(self, data, relocs, base_addr):
    if type(data) is not bytearray:
      data = bytearray(data)
    for off in relocs:
      addr = struct.unpack_from(">I", data, off)[0]
      if addr < base_addr:
        raise ValueError("Invalid relocatable address: %08x" % addr)
      addr -= base_addr
      struct.pack_into(">I", data, off, addr)
    return data

  def extract_bin_img(self, entry):
    data, relocs = self.extract_entry(entry)
    # create a bin image
    bi = BinImage(BIN_IMAGE_TYPE_HUNK)
    seg = Segment(SEGMENT_TYPE_CODE, len(data), data)
    bi.add_segment(seg)
    # create reloc for target segment
    rl = Relocations(seg)
    # add offsets
    for o in relocs:
      r = Reloc(o)
      rl.add_reloc(r)
    seg.add_reloc(seg, rl)
    # return final binary image
    return bi

  def write_index_file(self, idx_path):
    with open(idx_path, "w") as fh:
      for e in self.remus_rom.entries:
        fh.write(e.name + "\n")



import ctypes
import getpass
import logging
import os
import subprocess
import sys
import time
from functools import lru_cache
from typing import Any, Dict, List, Optional

# The original argument list at boot time, before any modifications
_argv = []  # type: List[str]

# The current directory when the application was started
_cwd = os.getcwd()

# Key-value store
_values: Dict[str, str] = {}

# Found Plugin/Python code override directory
_plugin_code_override = False

logger = logging.getLogger("BOOT")
perf_counter_epoch = time.perf_counter()


def init() -> None:
    global _argv, _cwd
    _cwd = os.getcwd()
    _argv = sys.argv.copy()

    setup_python_path()
    print("sys.path =", sys.path)


def set(key: str, value: str) -> None:
    _values[key] = value


def get(key: str, default: str = "") -> str:
    return _values.get(key, default)


def setup_logging() -> None:
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    # formatter = logging.Formatter(
    #     "[%(name)s] - %(asctime)s - %(levelname)s - %(message)s")
    formatter = logging.Formatter("[%(name)s] %(message)s")
    ch.setFormatter(formatter)
    root.addHandler(ch)


if "--workspace" in sys.argv:
    # Hack
    set("fws", "1")
if "--openretro" in sys.argv:
    # Hack
    set("openretro", "1")
if "--logging" in sys.argv:
    setup_logging()
    sys.argv.remove("--logging")


@lru_cache()
def executable_dir() -> str:
    """Returns the directory containing the executable (or main script)."""
    logger.debug("executable_dir")
    logger.debug("sys.executable = %s", sys.executable)
    if "python" in os.path.basename(sys.executable):
        logger.debug("Using sys.argv[0] instead of python interpreter path")
        # We do not want the directory of the (installed) python
        # interpreter, but rather the main application script.
        logger.debug("sys.argv[0] = %s", repr(sys.argv[0]))
        path = os.path.dirname(os.path.abspath(sys.argv[0]))
        logger.debug("%s", repr(path))
    else:
        path = os.path.dirname(sys.executable)
    if not os.path.isabs(path):
        logger.warning(
            "WARNING: executable_dir %s is not absolute", repr(path)
        )
        print(
            f"WARNING: executable_dir {repr(path)} is not absolute",
            file=sys.stderr,
        )
    logger.debug("executable_dir = %s", repr(path))
    return path


@lru_cache()
def app_dir() -> str:
    """Returns the (absolute) directory containing the application.

    This is the same as the executable_dir, except for macOS where the
    directory containing the .app bundle is returned.
    """
    app_dir = executable_dir()
    if is_macos() and is_frozen():
        # Break out of .app/Contents/MacOS
        app_dir = os.path.normpath(os.path.join(app_dir, "..", "..", ".."))
    return app_dir


@lru_cache()
def plugin_dir() -> Optional[str]:
    """Returns the (absolute) directory containing the plugin (or None)."""
    plugin_dir = os.path.join(app_dir(), "..", "..")
    if os.path.exists(os.path.join(plugin_dir, "Plugin.ini")):
        return plugin_dir
    return None


if sys.platform == "win32":
    S_OK = 0
    CSIDL_DESKTOP = 0
    CSIDL_PERSONAL = 5
    CSIDL_APPDATA = 26
    CSIDL_MYPICTURES = 39
    CSIDL_PROFILE = 40
    CSIDL_COMMON_DOCUMENTS = 46
    from ctypes import windll
    from ctypes import wintypes

    _SHGetFolderPath = windll.shell32.SHGetFolderPathW
    _SHGetFolderPath.argtypes = [
        wintypes.HWND,
        ctypes.c_int,
        wintypes.HANDLE,
        wintypes.DWORD,
        wintypes.LPCWSTR,
    ]

    def csidl_dir(csidl):
        path_buf = ctypes.create_unicode_buffer(wintypes.MAX_PATH)
        result = _SHGetFolderPath(0, csidl, 0, 0, path_buf)
        if result != S_OK:
            raise RuntimeError(
                "Could not find common directory for CSIDL {}".format(csidl)
            )
        return path_buf.value

    class WinPathsException(Exception):
        pass

    def err_unless_zero(result):
        if result == 0:
            return result
        else:
            raise WinPathsException(
                "Failed to retrieve windows path: %s" % result
            )

    _SHGetFolderPath.restype = err_unless_zero


@lru_cache()
def user_name() -> str:
    name = getpass.getuser()
    return name


@lru_cache()
def xdg_user_dir(name: str) -> Optional[str]:
    try:
        process = subprocess.Popen(
            ["xdg-user-dir", name], stdout=subprocess.PIPE
        )
        assert process.stdout is not None
        pathBytes = process.stdout.read().strip()
        path = pathBytes.decode("UTF-8")
        logger.debug("XDG user dir %s => %s", name, repr(path))
    except Exception:
        return None
    else:
        return path


@lru_cache()
def home_dir() -> str:
    if sys.platform == "win32":
        path = csidl_dir(CSIDL_PROFILE)
    else:
        path = os.path.expanduser("~")
    assert isinstance(path, str)
    # path = unicode_path(path)
    return path


@lru_cache()
def documents_dir(create: bool = False) -> str:
    if sys.platform == "win32":
        path = csidl_dir(CSIDL_PERSONAL)
    elif sys.platform == "darwin":
        path = os.path.join(home_dir(), "Documents")
    else:
        path = xdg_user_dir("DOCUMENTS")
        if not path:
            path = home_dir()
    assert isinstance(path, str)
    # path = unicode_path(path)
    if create and not os.path.isdir(path):
        os.makedirs(path)
    return path


def is_portable() -> bool:
    return portable_dir() is not None


@lru_cache()
def portable_dir() -> Optional[str]:
    path = executable_dir()
    last = ""
    while not last == path:
        portable_ini_path = os.path.join(path, "Portable.ini")
        logger.debug("Checking %s", repr(portable_ini_path))
        if os.path.exists(portable_ini_path):
            logger.debug("Detected portable dir: %s", repr(path))
            return path
        last = path
        path = os.path.dirname(path)
    logger.debug("No Portable.ini found in search path")
    return None


@lru_cache()
def common_data_dir(create: bool = False) -> str:
    if sys.platform == "win32":
        path = csidl_dir(CSIDL_APPDATA)
    elif sys.platform == "darwin":
        path = os.path.join(home_dir(), "Library", "Application Support")
    else:
        path = os.path.join(home_dir(), ".local", "share")
        path = os.environ.get("XDG_DATA_HOME", path)
    # path = unicode_path(path)
    if create and not os.path.exists(path):
        os.makedirs(path)
    return path


@lru_cache()
def app_data_dir(app: str) -> str:
    return os.path.join(common_data_dir(), app)


@lru_cache()
def app_config_dir(app: str) -> str:
    # if not app:
    #     app = get_app_id()
    if sys.platform == "win32":
        path = app_data_dir(app)
    elif sys.platform == "darwin":
        path = os.path.join(home_dir(), "Library", "Preferences", app)
    else:
        path = os.path.join(home_dir(), ".config")
        path = os.environ.get("XDG_CONFIG_HOME", path)
        path = os.path.join(path, app)
        # path = unicode_path(path)
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


@lru_cache()
def custom_path(name: str) -> Optional[str]:
    if get("openretro") == "1":
        app_names = ["openretro"]
    else:
        app_names = ["fs-uae-launcher", "fs-uae"]
    for app_name in app_names:
        key_path = os.path.join(app_config_dir(app_name), name)
        logger.debug("Checking %s", repr(key_path))
        if os.path.exists(key_path):
            try:
                with open(key_path, "r", encoding="UTF-8") as f:
                    path = f.read().strip()
                    break
            except Exception as e:
                logger.debug("Error reading custom path %s", repr(e))
    else:
        return None
    path_lower = path.lower()
    if path_lower.startswith("$home/") or path_lower.startswith("$home\\"):
        path = os.path.join(home_dir(), path[6:])
    elif path_lower.startswith("~/") or path_lower.startswith("~\\"):
        path = os.path.join(home_dir(), path[2:])
    print(f"custom_path {name!r} -> {path!r}")
    logger.debug("custom_path %r -> %r", name, path)
    return path


def _getBaseDirectoryBeforeNormalization() -> str:
    path: Optional[str]
    logger.debug("Find base_dir")
    for arg in sys.argv[1:]:
        if arg.startswith("--base-dir="):
            path = arg[11:]
            path = os.path.abspath(path)
            logger.debug("Using base_dir via argv: %r", path)
            sys.argv.remove(arg)
            return path

    path = portable_dir()
    if path:
        return path

    if get("openretro") == "1":
        logger.debug("Checking OPENRETRO_BASE_DIR")
        path = os.environ.get("OPENRETRO_BASE_DIR", "")
        if path:
            logger.debug("Using base_dir via OPENRETRO_BASE_DIR: %r", path)
            return path

    else:
        logger.debug("Checking FS_UAE_BASE_DIR")
        path = os.environ.get("FS_UAE_BASE_DIR", "")
        if path:
            logger.debug("Using base_dir via FS_UAE_BASE_DIR: %r", path)
            return path

    path = custom_path("base-dir")
    if path:
        logger.debug("Using base_dir via custom path config: %r", path)
        return path

    if get("base_dir_name"):
        # path = os.path.join(documents_dir(True), get("base_dir_name"))
        path = os.path.join(home_dir(), get("base_dir_name"))
    elif get("openretro") == "1":
        path = os.path.join(home_dir(), "OpenRetro")
    else:
        # Check new ~/FS-UAE directory first
        path = os.path.join(home_dir(), "FS-UAE")
        if os.path.exists(path):
            print("- using base_dir $HOME/FS-UAE")
        else:
            print("- using base_dir $DOCUMENTS/FS-UAE")
            # FS-UAE uses Documents/FS-UAE for legacy reasons
            path = os.path.join(documents_dir(), "FS-UAE")

    assert path is not None
    logger.debug("Using default base_dir %r", path)
    return path


@lru_cache()
def base_dir() -> str:
    path = _getBaseDirectoryBeforeNormalization()
    print("Using base_dir =", repr(path))
    # It is important to remove any trailing slash, etc. since the exact
    # base_dir is used for removing base_dir prefix from other paths.
    # Case normalization is important too, but that's done by FSGSDirectories.
    path = os.path.normpath(path)
    print("After normpath: base_dir =", repr(path))
    logger.debug("After normpath, base_dir = %r", path)
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def getBaseDirectory() -> str:
    return base_dir()


@lru_cache()
def development() -> bool:
    # FIXME: Document option
    if "--development-mode=0" in sys.argv:
        return False
    result = os.path.exists(os.path.join(executable_dir(), "pyproject.toml"))
    logger.info("Development mode: %s", result)
    return result


def is_frozen() -> bool:
    return bool(getattr(sys, "frozen", False))


def is_macos() -> bool:
    return sys.platform == "darwin"


def setup_frozen_python_libs() -> None:
    return

    # libs_dir = os.path.abspath(
    #     os.path.join(executable_dir(), "..", "..", "Python")

    # libs_dirs = [executable_dir()]
    # if sys.platform == "darwin":
    #     # Add .app/Contents/Python to libs_dirs
    #     libs_dir = os.path.abspath(
    #         os.path.join(executable_dir(), "..", "Python")
    #     )
    #     print(libs_dir, os.path.exists(libs_dir))
    #     if os.path.exists(libs_dir):
    #         libs_dirs.append(libs_dir)
    # libs_dir = os.path.abspath(
    #     os.path.join(executable_dir(), "..", "..", "Python")
    # )
    # print(libs_dir, os.path.exists(libs_dir))
    # if os.path.exists(libs_dir):
    #     libs_dirs.append(libs_dir)
    # else:
    #     libs_dir = os.path.abspath(
    #         os.path.join(
    #             executable_dir(), "..", "..", "..", "..", "..", "Python"
    #         )
    #     )
    #     print(libs_dir, os.path.exists(libs_dir))
    #     if os.path.exists(libs_dir):
    #         libs_dirs.append(libs_dir)
    # for libs_dir in libs_dirs:
    #     for item in os.listdir(libs_dir):
    #         if item.endswith(".zip"):
    #             path = os.path.join(libs_dir, item)
    #             print("adding", path)
    #             sys.path.insert(0, path)


def plugin_code_override() -> bool:
    return _plugin_code_override


# from importlib.abc import MetaPathFinder

# _python_dir: str


# class OverrideImporter(MetaPathFinder):

#     # def find_spec(self, fullname, path, target=None):
#     #     from importlib.util import spec_from_loader
#     #     loader = self.find_module(fullname, path)
#     #     if loader is None:
#     #         return None
#     #     spec = spec_from_loader(fullname, loader)
#     #     print(spec)
#     #     return spec

#     # FIXME: Any
#     def find_module(self, fullname: str, path: str) -> Any:
#         from importlib.machinery import SourceFileLoader

#         print("CustomImporter.find_module", fullname, path)
#         # if path is None:
#         #     raise Exception("")
#         if path is None:
#             package_path = os.path.join(_python_dir, fullname, "__init__.py")
#             print(package_path)
#             if os.path.exists(package_path):
#                 print("->", package_path)
#                 return SourceFileLoader(fullname, package_path)
#             # module_path = os.path.join(_python_dir, f"{fullname}.py")
#             # if os.path.exists(module_path):
#             #     return SourceFileLoader(fullname, module_path)
#         else:
#             if len(path) == 0:
#                 return None
#             name = fullname.rsplit(".", 1)[-1]
#             package_path = os.path.join(path[0], name, "__init__.py")
#             print(package_path)
#             if os.path.exists(package_path):
#                 print("->", package_path)
#                 return SourceFileLoader(fullname, package_path)
#             module_path = os.path.join(path[0], f"{name}.py")
#             print(module_path)
#             if os.path.exists(module_path):
#                 print("->", module_path)
#                 return SourceFileLoader(fullname, module_path)
#         print("->", None)
#         return None


def setup_python_path_frozen() -> None:
    """Allow overriding code from Plugin/Python directory."""
    print("setup_python_path_frozen")
    # global _python_dir
    # base_dir = plugin_dir() or app_dir()
    # python_dir = os.path.join(base_dir, "Python")
    # # print("Check", python_dir)
    # # if os.path.exists(python_dir):
    # #     # sys.path.insert(0, python_dir)
    # #     # _plugin_code_override = True
    # #     _python_dir = python_dir
    # #     sys.meta_path.insert(0, OverrideImporter())
    # #     print("sys.meta_path", sys.meta_path)


def setup_python_path() -> None:
    # print("setup_python_path")
    if is_frozen():
        setup_python_path_frozen()



import fsui
from launcher.i18n import gettext
from launcher.setup.setupwizardpage import SetupWizardPage
from launcher.setup.wizardheader import WizardHeader


class SetupWelcomePage(SetupWizardPage):
    def __init__(self, parent):
        super().__init__(parent)
        self.layout = fsui.VerticalLayout()

        header = WizardHeader(
            self, fsui.Icon("fs-uae-launcher", "pkg:launcher"), "Welcome"
        )
        self.layout.add(header, fill=True, margin_bottom=20)

        v_layout = fsui.VerticalLayout()
        self.layout.add(v_layout, margin=20)

        label = fsui.MultiLineLabel(
            self,
            gettext("Welcome to the setup wizard for FS-UAE Launcher!"),
            SetupWizardPage.WIDTH,
        )
        v_layout.add(label, fill=True, margin_top=0)

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "You can close this wizard at any time if you do not want to "
                "use it, and you can also run it again later."
            ),
            SetupWizardPage.WIDTH,
        )
        v_layout.add(label, fill=True, margin_top=20)



from functools import lru_cache

from fsgs.drivers.gamedriver import GameDriver


class PlatformHandler:
    def __init__(self, loader_class=None, runner_class=None):
        self.loader_class = loader_class
        self.runner_class = runner_class

    @classmethod
    def create(cls, platform_id):
        class_ = cls.get_platform_class(platform_id)
        return class_()

    @classmethod
    def get_platform_class(cls, platform_id):
        platform_id = platform_id.lower()
        try:
            return platforms[platform_id]
        except KeyError:
            return UnsupportedPlatform

    @classmethod
    @lru_cache()
    def get_platform_name(cls, platform_id):
        return cls.get_platform_class(platform_id).PLATFORM_NAME

    @classmethod
    @lru_cache()
    def get_platform_ids(cls):
        return sorted(platforms.keys())

    def get_loader(self, fsgs):
        if self.loader_class is None:
            raise Exception("loader class is None for " + repr(self))
        return self.loader_class(fsgs)

    def get_runner(self, fsgs) -> GameDriver:
        if self.runner_class is None:
            raise Exception("runner class is None for " + repr(self))
        return self.runner_class(fsgs)

    def driver(self, fsgc):
        return self.get_runner(fsgc)

    def loader(self, fsgc):
        return self.get_loader(fsgc)


class Platform(PlatformHandler):
    AMIGA = "amiga"
    ARCADE = "arcade"
    A2600 = "a2600"
    A5200 = "a5200"
    A7800 = "a7800"
    ATARI = "atari"
    C64 = "c64"
    CD32 = "cd32"
    CDTV = "cdtv"
    CPC = "cpc"
    DOS = "dos"
    GB = "gb"
    GBA = "gba"
    GBC = "gbc"
    LYNX = "lynx"
    MSX = "msx"
    N64 = "n64"
    NDS = "nds"
    NEOGEO = "neogeo"
    NES = "nes"
    NGC = "ngc"
    PSX = "psx"
    SNES = "snes"
    SGG = "sgg"
    SMD = "smd"
    SMS = "sms"
    TG16 = "tg16"
    TGCD = "tgcd"
    ZXS = "zxs"

    def get_loader(self, fsgc):
        return self.loader(fsgc)

    def get_runner(self, fsgc):
        return self.driver(fsgc)

    def driver(self, fsgc):
        raise NotImplementedError()

    def loader(self, fsgc):
        raise NotImplementedError()


from fsgs.platforms.amiga import AmigaPlatformHandler
from fsgs.platforms.amstrad_cpc import AmstradCPCPlatformHandler
from fsgs.platforms.arcade.arcadeplatform import ArcadePlatformHandler
from fsgs.platforms.atari_2600 import Atari2600PlatformHandler
from fsgs.platforms.atari5200 import Atari5200PlatformHandler
from fsgs.platforms.atari7800 import Atari7800PlatformHandler
from fsgs.platforms.atari.atariplatform import AtariSTPlatformHandler
from fsgs.platforms.cd32 import CD32PlatformHandler
from fsgs.platforms.cdtv import CDTVPlatformHandler
from fsgs.platforms.commodore64 import Commodore64Platform
from fsgs.platforms.dos.dosplatform import DOSPlatformHandler
from fsgs.platforms.gameboy import GameBoyPlatform
from fsgs.platforms.gameboyadvance import GameBoyAdvancePlatform
from fsgs.platforms.gameboycolor import GameBoyColorPlatform
from fsgs.platforms.gamegear import GameGearPlatform
from fsgs.platforms.lynx import LynxPlatformHandler
from fsgs.platforms.mastersystem import MasterSystemPlatform
from fsgs.platforms.megadrive import MegaDrivePlatform
from fsgs.platforms.msx import MsxPlatformHandler
from fsgs.platforms.nintendo64 import Nintendo64Platform
from fsgs.platforms.nintendods import NintendoDSPlatform
from fsgs.platforms.neogeo import NeoGeoPlatform
from fsgs.platforms.gamecube import GameCubePlatform
from fsgs.platforms.nintendo import NintendoPlatform
from fsgs.platforms.playstation import PlayStationPlatform
from fsgs.platforms.supernintendo import SuperNintendoPlatformHandler
from fsgs.platforms.turbografx16 import TurboGrafx16Platform
from fsgs.platforms.turbografxcd import TurboGrafxCDPlatform
from fsgs.platforms.zxs import SpectrumPlatformHandler


class UnsupportedPlatform(PlatformHandler):
    PLATFORM_NAME = "Unsupported"


platforms = {
    Platform.AMIGA: AmigaPlatformHandler,
    Platform.ARCADE: ArcadePlatformHandler,
    Platform.A2600: Atari2600PlatformHandler,
    Platform.A5200: Atari5200PlatformHandler,
    Platform.A7800: Atari7800PlatformHandler,
    Platform.ATARI: AtariSTPlatformHandler,
    Platform.C64: Commodore64Platform,
    Platform.CD32: CD32PlatformHandler,
    Platform.CDTV: CDTVPlatformHandler,
    Platform.CPC: AmstradCPCPlatformHandler,
    Platform.DOS: DOSPlatformHandler,
    Platform.GB: GameBoyPlatform,
    Platform.GBA: GameBoyAdvancePlatform,
    Platform.GBC: GameBoyColorPlatform,
    Platform.LYNX: LynxPlatformHandler,
    Platform.MSX: MsxPlatformHandler,
    Platform.N64: Nintendo64Platform,
    Platform.NDS: NintendoDSPlatform,
    Platform.NEOGEO: NeoGeoPlatform,
    Platform.NES: NintendoPlatform,
    Platform.NGC: GameCubePlatform,
    Platform.SNES: SuperNintendoPlatformHandler,
    Platform.PSX: PlayStationPlatform,
    Platform.SGG: GameGearPlatform,
    Platform.SMD: MegaDrivePlatform,
    Platform.SMS: MasterSystemPlatform,
    Platform.TG16: TurboGrafx16Platform,
    Platform.TGCD: TurboGrafxCDPlatform,
    Platform.ZXS: SpectrumPlatformHandler,
}
PLATFORM_IDS = platforms.keys()


def normalize_platform_id(platform_id):
    platform_id = platform_id.lower().replace("-", "").replace("_", "")
    # noinspection SpellCheckingInspection
    if platform_id in ["st", "atarist"]:
        return Platform.ATARI
    elif platform_id in ["commodorecdtv"]:
        return Platform.CDTV
    elif platform_id in ["amigacd32"]:
        return Platform.CD32
    elif platform_id in ["amstradcpc"]:
        return Platform.CPC
    elif platform_id in ["msdos"]:
        return Platform.DOS
    elif platform_id in ["gameboy"]:
        return Platform.GB
    elif platform_id in ["gameboyadvance"]:
        return Platform.GBA
    elif platform_id in ["gameboycolor"]:
        return Platform.GBC
    elif platform_id in ["nintendo", "famicom"]:
        return Platform.NES
    elif platform_id in ["nintendo64"]:
        return Platform.N64
    elif platform_id in ["nintendods"]:
        return Platform.NDS
    elif platform_id in ["nintendogamecube", "gamecube", "gc"]:
        return Platform.NGC
    elif platform_id in ["supernintendo", "supernes", "superfamicom"]:
        return Platform.SNES
    elif platform_id in ["zxspectrum"]:
        return Platform.ZXS
    elif platform_id in ["mastersystem", "segamastersystem"]:
        return Platform.SMS
    elif platform_id in ["megadrive", "segamegadrive", "genesis"]:
        return Platform.SMD
    elif platform_id in ["atari2600"]:
        return Platform.A2600
    elif platform_id in ["atari5200"]:
        return Platform.A5200
    elif platform_id in ["atari78600"]:
        return Platform.A7800
    elif platform_id in ["turbografx16"]:
        return Platform.TG16
    elif platform_id in ["gamegear", "sgg"]:
        return Platform.SGG
    return platform_id



from __future__ import absolute_import
from __future__ import print_function

import time
from .TimeStamp import *

class RootMetaInfo:
  def __init__(self, create_ts=None, disk_ts=None, mod_ts=None):
    self.set_create_ts(create_ts)
    self.set_disk_ts(disk_ts)
    self.set_mod_ts(mod_ts)
  
  def __str__(self):
    res = []
    res.append(self.get_create_time_str())
    res.append(self.get_disk_time_str())
    res.append(self.get_mod_time_str())
    return "  ".join(res)
      
  # create_ts
  def set_create_time(self, create_time):
    self.create_time = create_time
    if self.create_time != None:
      self.create_ts = TimeStamp()
      self.create_ts.from_secs(create_time)
    else:
      self.create_ts = None
  
  def set_create_ts(self, create_ts):
    self.create_ts = create_ts
    if self.create_ts != None:
      self.create_time = self.create_ts.get_secsf()
    else:
      self.create_time = None
  
  def get_create_time(self):
    return self.create_time
  
  def get_create_ts(self):
    return self.create_ts

  def get_create_time_str(self):
    if self.create_ts != None:
      return str(self.create_ts)
    else:
      return ts_empty_string

  def set_current_as_create_time(self):
    create_time = time.mktime(time.localtime())
    self.set_create_time(create_time)

  # disk_ts
  def set_disk_time(self, disk_time):
    self.disk_time = disk_time
    if self.disk_time != None:
      self.disk_ts = TimeStamp()
      self.disk_ts.from_secs(disk_time)
    else:
      self.disk_ts = None
  
  def set_disk_ts(self, disk_ts):
    self.disk_ts = disk_ts
    if self.disk_ts != None:
      self.disk_time = self.disk_ts.get_secsf()
    else:
      self.disk_time = None
  
  def get_disk_time(self):
    return self.disk_time
  
  def get_disk_ts(self):
    return self.disk_ts

  def get_disk_time_str(self):
    if self.disk_ts != None:
      return str(self.disk_ts)
    else:
      return ts_empty_string

  def set_current_as_disk_time(self):
    disk_time = time.mktime(time.localtime())
    self.set_disk_time(disk_time)

  # mod_ts
  def set_mod_time(self, mod_time):
    self.mod_time = mod_time
    if self.mod_time != None:
      self.mod_ts = TimeStamp()
      self.mod_ts.from_secs(mod_time)
    else:
      self.mod_ts = None
  
  def set_mod_ts(self, mod_ts):
    self.mod_ts = mod_ts
    if self.mod_ts != None:
      self.mod_time = self.mod_ts.get_secsf()
    else:
      self.mod_time = None
  
  def get_mod_time(self):
    return self.mod_time
  
  def get_mod_ts(self):
    return self.mod_ts
  
  def get_mod_time_str(self):
    if self.mod_ts != None:
      return str(self.mod_ts)
    else:
      return ts_empty_string
 
  def set_current_as_mod_time(self):
    mod_time = time.mktime(time.localtime())
    self.set_mod_time(mod_time)

  


import weakref
from launcher.launcher_settings import LauncherSettings


class SettingsBehavior:
    def __init__(self, parent, names):
        parent.__settings_enable_behavior = self
        self._parent = weakref.ref(parent)
        self._names = set(names)
        LauncherSettings.add_listener(self)
        try:
            parent.destroyed.connect(self.on_parent_destroyed)
        except AttributeError:
            print(
                "WARNING: SettingsBehavior without remove_listener "
                "implementation"
            )
        for name in names:
            # Broadcast initial value
            self.on_setting(name, LauncherSettings.get(name))

    def on_parent_destroyed(self):
        print("SettingsBehavior: remove_listener", self._parent())
        LauncherSettings.remove_listener(self)

    def on_setting(self, key, value):
        if key in self._names:
            widget = self._parent()
            try:
                func = getattr(widget, "on_{0}_setting".format(key))
            except AttributeError:
                func = getattr(widget, "on_settings".format(key))
                func(key, value)
            else:
                func(value)



import fsui
from fsui.qt import QFont


class Label(fsui.Label):
    def __init__(self, parent, text="", font=None):
        super().__init__(parent, text)
        if font is not None:
            if isinstance(font, str):
                size, weight, family = font.split(" ", 3)
                assert size.endswith("px")
                font = QFont(family)
                font.setPixelSize(int(size[:-2]))
                weight = weight.lower()
                if weight == "regular":
                    font.setWeight(QFont.Normal)
                elif weight == "medium":
                    font.setWeight(QFont.DemiBold)
                elif weight == "bold":
                    font.setWeight(QFont.Bold)
                else:
                    raise Exception("Unknown weight")
                font = fsui.Font(font)

            self.set_font(font)


class MultiLineLabel(fsui.MultiLineLabel):
    pass



import io
import os
import shutil
import tempfile
import traceback
import warnings
from collections import defaultdict

import fsboot
from fsbc.application import Application
from fsbc.resources import Resources
from fsbc.settings import Settings
from fsbc.system import System
from fsbc.task import current_task
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.fsuae import FSUAE
from launcher.option import Option
from fsgs.plugins.pluginexecutablefinder import PluginExecutableFinder
from fsgs.plugins.pluginmanager import PluginManager, Executable
from fsgs.refreshratetool import RefreshRateTool
from fsgs.util.gamenameutil import GameNameUtil


class GameDriverLogger:
    def debug(self, message):
        print("[DEBUG]", message)

    def info(self, message):
        print("[INFO]", message)

    def warning(self, message):
        # FIXME: Store warning and show in some kind of launch log
        # (i.e. "warnings and error generated during game execution")
        print("[WARNING]", message)

    def error(self, message):
        print("[ERROR]", message)


# noinspection PyMethodMayBeStatic
class GameDriver:
    PORTS = []

    def __init__(self, fsgc):
        self.fsgc = fsgc
        self.files = GameFiles(fsgc)
        self.logger = GameDriverLogger()
        self.emulator = Emulator("no-emulator")

        self.options = defaultdict(str)
        self.init_options()

        self.ports = []
        self.init_ports()

        self._allow_gsync = True
        self._model_name = ""
        self._emulator_skin_prepared = {}
        self.__vsync = False
        self.__game_temp_file = None

        self.temp_root = TemporaryItem(
            root=None, prefix="fsgs-", suffix="tmp", directory=True
        )
        # # Default current working directory for the emulator.
        self.cwd = self.temp_dir("cwd")
        # # Fake home directory for the emulator.
        self.home = self.temp_dir("home")
        self.home._path = os.path.join(FSGSDirectories.get_cache_dir(), "Home")
        if not os.path.exists(self.home.path):
            os.makedirs(self.home.path)
        # noinspection PyProtectedMember
        self.cwd._path = self.home._path

    def __del__(self):
        print("GameDriver.__del__")

    def prepare(self):
        pass

    def install(self):
        self.files.install()

    def run(self):
        executable = PluginExecutableFinder().find_executable(self.emulator.name)
        if executable is None:
            raise LookupError(
                "Could not find emulator " + repr(self.emulator.name)
            )
        self.emulator.process = self.start_emulator(Executable(executable))

    def emulator_pid_file(self):
        return self.fsgc.settings[Option.EMULATOR_PID_FILE]

    @classmethod
    def write_emulator_pid_file(cls, pid_file, process):
        if pid_file:
            try:
                with open(pid_file, "w") as f:
                    f.write("{}".format(process.pid))
            except Exception:
                traceback.print_exc()
                print("WARNING: Error writing emulator PID file")

    @classmethod
    def remove_emulator_pid_file(cls, pid_file):
        if not pid_file:
            return
        try:
            os.remove(pid_file)
        except Exception:
            traceback.print_exc()
            print("WARNING: Error removing emulator PID file")

    def wait(self):
        result = self.emulator.process.wait()
        self.remove_emulator_pid_file(self.emulator_pid_file())
        return result

    def finish(self):
        pass

    def init_options(self):
        for key, value in Settings.instance().values.items():
            # FIXME: re-enable this check?
            # if key in Config.config_keys:
            #     print("... ignoring config key from settings:", key)
            #     continue
            self.options[key] = value
        for key, value in self.fsgc.config.items():
            self.options[key] = value

    def init_ports(self):
        for i, port_info in enumerate(self.PORTS):
            self.init_port(i, port_info)

    def init_port(self, index, port_info):
        port = Port(port_info["description"])
        # FIXME: MOVE TO Port CONSTRUCTOR
        port.number = index + 1
        port.types = port_info["types"]
        port.type_option = port_info.get("type_option", "")
        port.device_option = port_info.get("device_option", "")
        # Set correct (port type index) based on type option
        # FIXME: MOVE CODE TO port CLASS
        if port.type_option:
            type_value = self.options[port.type_option]
            if type_value:
                print(
                    "[INPUT] Port option",
                    port.type_option,
                    "was set to",
                    type_value,
                )
            else:
                print("[INPUT] No value specified for", port.type_option)

            # Hack / Workaround for Amiga
            # FIXME: Move to driver subclasses?
            # if not type_value:
            #     if self.options[Option.AMIGA_MODEL].lower() == "cd32":
            #         type_value = "cd32 pad"

            if not type_value:
                # FIXME: RETRIEVE DEFAULT VALUE FOR TYPE OPTION
                try:
                    type_value = Option.get(port.type_option)["default"]
                except Exception:
                    print(
                        "[INPUT] Could not find default for", port.type_option
                    )
            for j, port_type in enumerate(port.types):
                print("[INPUT]", j, port_type["type"])
                if type_value == port_type["type"]:
                    print("[INPUT] Set index", j, "for type", type_value)
                    port.index = j
                    break
            else:
                print(
                    "[INPUT] WARNING: Could not find index for type",
                    type_value,
                )
        self.ports.append(port)

    def set_allow_gsync(self, allow):
        self._allow_gsync = False

    def set_model_name(self, name):
        self._model_name = name

    def use_fullscreen(self):
        # FIXME: not a very nice hack to hard-code application name here...
        if Application.instance():
            if Application.instance().name == "fs-uae-arcade":
                return True
        if Settings.instance()["fullscreen"] == "0":
            return False
        return True

    def fullscreen_window_mode(self):
        fullscreen_mode = self.options[Option.FULLSCREEN_MODE]
        return fullscreen_mode == "window"

    def g_sync(self):
        return self.options[Option.G_SYNC] == "1"

    def use_vsync(self):
        if self.g_sync():
            return False
        # if "--no-vsync" in sys.argv:
        #     return False
        # return True
        return self.options[Option.VIDEO_SYNC] in ["1", "auto"]

    def use_doubling(self):
        return True

    NO_SMOOTHING = "0"
    SMOOTHING = "1"
    NON_INTEGER_SMOOTHING = "auto"
    DEFAULT_SMOOTHING = NON_INTEGER_SMOOTHING

    def smoothing(self):
        if self.options[Option.SMOOTHING] == self.NO_SMOOTHING:
            return self.NO_SMOOTHING
        if self.options[Option.SMOOTHING] == self.SMOOTHING:
            return self.SMOOTHING
        if self.options[Option.SMOOTHING] == self.NON_INTEGER_SMOOTHING:
            return self.NON_INTEGER_SMOOTHING
        return self.DEFAULT_SMOOTHING

    NO_SCALING = "0"
    MAX_SCALING = "1"
    INTEGER_SCALING = "integer"
    DEFAULT_SCALING = MAX_SCALING

    def scaling(self):
        if self.options[Option.SCALE] == self.NO_SCALING:
            return self.NO_SCALING
        if self.options[Option.SCALE] == self.MAX_SCALING:
            return self.MAX_SCALING
        if self.options[Option.SCALE] == self.INTEGER_SCALING:
            return self.INTEGER_SCALING
        return self.DEFAULT_SCALING

    NO_STRETCHING = "0"
    STRETCH_FILL_SCREEN = "1"
    STRETCH_ASPECT = "aspect"
    DEFAULT_STRETCHING = STRETCH_ASPECT

    def stretching(self):
        if self.options[Option.STRETCH] == self.NO_STRETCHING:
            return self.NO_STRETCHING
        if self.options[Option.STRETCH] == self.STRETCH_FILL_SCREEN:
            return self.STRETCH_FILL_SCREEN
        if self.options[Option.STRETCH] == self.STRETCH_ASPECT:
            return self.STRETCH_ASPECT
        return self.DEFAULT_STRETCHING

    NO_BORDER = "0"
    SMALL_BORDER = "1"
    LARGE_BORDER = "large"
    DEFAULT_BORDER = SMALL_BORDER

    def border(self):
        if self.options[Option.BORDER] == self.NO_BORDER:
            return self.NO_BORDER
        if self.options[Option.BORDER] == self.SMALL_BORDER:
            return self.SMALL_BORDER
        if self.options[Option.BORDER] == self.LARGE_BORDER:
            return self.LARGE_BORDER
        return self.DEFAULT_BORDER

    NO_EFFECT = "0"
    CRT_EFFECT = "crt"
    HQ2X_EFFECT = "hq2x"
    SCALE2X_EFFECT = "scale2x"
    DOUBLE_EFFECT = "2x"
    DEFAULT_EFFECT = DOUBLE_EFFECT

    def effect(self):
        if self.options[Option.EFFECT] == self.NO_EFFECT:
            return self.NO_EFFECT
        if self.options[Option.EFFECT] == self.CRT_EFFECT:
            return self.CRT_EFFECT
        if self.options[Option.EFFECT] == self.HQ2X_EFFECT:
            return self.HQ2X_EFFECT
        if self.options[Option.EFFECT] == self.SCALE2X_EFFECT:
            return self.SCALE2X_EFFECT
        return self.DEFAULT_EFFECT

    def bezel(self):
        return self.options[Option.BEZEL] == "1"

    # FIXME: REMOVE
    def use_smoothing(self):
        return True

    # FIXME: REMOVE
    def use_stretching(self):
        if self.options[Option.KEEP_ASPECT] == "1":
            return
        # if "--no-stretch" in sys.argv:
        #     return False
        # if self.options[Option.KEEP_ASPECT] == "1":
        #     return
        # if Settings.instance()["stretch"] == "0":
        #     return False
        return True

    def use_auto_load(self):
        if self.options[Option.AUTO_LOAD] == "0":
            return False
        return True

    def use_turbo_load(self):
        if self.options[Option.TURBO_LOAD] == "0":
            return False
        return True

    def use_audio_frequency(self):
        if self.options[Option.AUDIO_FREQUENCY]:
            try:
                return int(Settings.instance()[Option.AUDIO_FREQUENCY])
            except ValueError:
                pass
        return 48000

    def screenshots_dir(self):
        return FSGSDirectories.screenshots_output_dir()

    def screenshots_base(self):
        return GameNameUtil.create_link_name(self.game_name())

    def screenshots_name(self):
        # FIXME: REMOVE?
        return GameNameUtil.create_fs_name(self.game_name())

    # def get_game_name(self):
    #     return self.options["game_name"]
    #
    # def get_variant_name(self):
    #     return self.options["variant_name"]

    def get_platform_name(self):
        # FIXME: UNUSED? REMOVE?
        p = self.options[Option.PLATFORM].lower()
        if p == "atari-7800":
            return "Atari 7800"
        if p == "amiga":
            return "Amiga"
        if p == "cdtv":
            return "CDTV"
        if p == "cd32":
            return "CD32"
        raise Exception("Unrecognized platform")

    def screen_size(self):
        rect = self.screen_rect()
        return rect[2], rect[3]

    @classmethod
    def _screens(cls):
        screens = []
        try:
            from fsui.qt import init_qt

            qapplication = init_qt()
            desktop = qapplication.desktop()
        except AttributeError:
            # no QApplication, probably not running via QT
            # FIXME: log warning
            pass
        else:
            for i in range(desktop.screenCount()):
                geometry = desktop.screenGeometry(i)
                screens.append([geometry.x(), i, geometry])
        return screens

    @classmethod
    def screen_refresh_rate_for_monitor(cls, monitor):
        try:
            from fsui.qt import init_qt

            qapplication = init_qt()
        except AttributeError:
            return 0
        else:
            for i, screen in enumerate(qapplication.screens()):
                print(
                    "Screen {0} refresh rate (Qt) = {1}".format(
                        i, screen.refreshRate()
                    )
                )
            index = cls.screen_index_for_monitor(monitor)
            screen = qapplication.screens()[index]

            if System.windows or System.macos:
                return int(round(screen.refreshRate()))

            refresh_rate_tool = RefreshRateTool()
            screens = refresh_rate_tool.screens_xrandr()
            rect = cls.screen_rect_for_monitor(monitor)
            for screen in screens:
                print(
                    "Screen {} refresh rate (Xrandr) = {}".format(
                        screen, screens[screen]["refresh_rate"]
                    )
                )
            for screen in screens:
                if rect == screen:
                    return screens[screen]["refresh_rate"]
            return 0

    def screen_refresh_rate(self):
        monitor = self.options[Option.MONITOR]
        return self.screen_refresh_rate_for_monitor(monitor)

    @classmethod
    def screen_index_for_monitor(cls, monitor):
        rect = cls.screen_rect_for_monitor(monitor)
        for i, s in enumerate(cls._screens()):
            if rect == (s[2].x(), s[2].y(), s[2].width(), s[2].height()):
                return i
        raise Exception("Could not find screen at position {}".format(rect))

    def screen_index(self):
        monitor = self.options[Option.MONITOR]
        return self.screen_index_for_monitor(monitor)

    @classmethod
    def screen_rect_for_monitor(cls, monitor):
        screens = cls._screens()
        screens.sort()
        if monitor == "left":
            mon = 0
        elif monitor == "middle-right":
            mon = 2
        elif monitor == "right":
            mon = 3
        else:  # middle-left
            mon = 1
        display = round(mon / 3 * (len(screens) - 1))
        geometry = screens[display][2]
        return geometry.x(), geometry.y(), geometry.width(), geometry.height()

    def screen_rect(self):
        monitor = self.options[Option.MONITOR]
        return self.screen_rect_for_monitor(monitor)

    def game_name(self):
        if self.fsgc.game.name:
            return self.fsgc.game.name
        if self.options["game_name"]:
            return self.options["game_name"]
        if self.options[Option.CONFIG_NAME]:
            return self.options[Option.CONFIG_NAME]
        return "Unknown Game"

    def variant_name(self):
        if self.fsgc.game.variant.name:
            return self.fsgc.game.variant.name
        if self.options["variant_name"]:
            return self.options["variant_name"]
        return "Unknown Variant"

    def get_name(self):
        # return "{0} ({1}, {2})".format(
        #     self.get_game_name(), self.get_platform_name(),
        #     self.get_variant_name())
        return "{0} ({1}, {2})".format(
            self.fsgc.game.name,
            self.fsgc.game.platform.name,
            self.fsgc.game.variant.name,
        )

    def save_dir(self, create=False):
        """Uses the new saves layout."""
        uuid = self.fsgc.game.variant.uuid
        if uuid:
            path = os.path.join(
                FSGSDirectories.saves_dir, "UUID", uuid[:3], uuid
            )
            if create and not os.path.exists(path):
                os.makedirs(path)
            return path
        # For now, return old save state dirs for non-UUID variants
        return self.get_state_dir()

    def get_state_dir(self):
        """This is the old deprecated save sate directory for games."""
        state_dir = os.path.join(
            FSGSDirectories.get_save_states_dir(),
            GameNameUtil.create_fs_name(self.get_name()),
        )
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        if self.fsgc.game.variant.uuid:
            uuid_file = os.path.join(state_dir, "uuid.txt")
            with io.open(uuid_file, "w", encoding="UTF-8") as f:
                f.write(self.fsgc.game.variant.uuid + "\n")
        # timestamp_file = os.path.join(state_dir, "timestamp.txt")
        # with open(timestamp_file, "wb") as f:
        #     f.write("\n")
        return state_dir

    def emulator_state_dir(self, emulator):
        assert emulator
        state_dir = os.path.join(self.get_state_dir(), emulator)
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        return state_dir

    def get_game_file(self, config_key="cartridge_slot"):
        if self.__game_temp_file is not None:
            return self.__game_temp_file.path

        file_uri = self.options[config_key]
        # FIXME: create new API fsgc.file.path(file_uri), returns path
        # (and creates temp file if necessary)
        input_stream = self.fsgc.file.open(file_uri)

        # self.__game_temp_file = self.fsgc.temp_file(file_uri.split("/")[-1])
        self.__game_temp_file = TemporaryNamedItem(
            self.temp_root, file_uri.split("/")[-1]
        )

        with open(self.__game_temp_file.path, "wb") as f:
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
        return self.__game_temp_file.path

    def temp_dir(self, name):
        return TemporaryNamedItem(
            root=self.temp_root, name=name, directory=True
        )

    def temp_file(self, name):
        return TemporaryNamedItem(
            root=self.temp_root, name=name, directory=False
        )

    def start_emulator_from_plugin_resource(
        self, provide_name, args=None, env_vars=None
    ):
        resource = self.fsgc.plugins.find_executable(provide_name)
        if resource is None:
            raise Exception("Could not find emulator " + repr(provide_name))
        if env_vars is None:
            env_vars = os.environ.copy()
        # Set LD_LIBRARY_PATH for Linux plugins with bundled libraries
        # EDIT: No longer doing, can cause problems for OpenGL drivers.
        # Instead, use the standalone-linux.py script with rpath fixing.
        # env_vars["LD_LIBRARY_PATH"] = os.path.dirname(resource.path)
        return self.start_emulator(
            "", args=args, env_vars=env_vars, executable=resource.path
        )

    def set_environment_path(self, env, key, value):
        """Supports exporting paths with MBCS encoding as well on Windows."""
        env[key] = value
        # if System.windows:
        #     env[key + "_MBCS"] = value.encode("mbcs")

    def update_environment(self, env):
        x, y, w, h = self.screen_rect()
        if env.get("FSGS_FULLSCREEN_RECT", ""):
            # Already specified externally, so we just use
            # the existing values...
            print("using existing fullscreen rect")
        else:
            env["FSGS_FULLSCREEN_RECT"] = "{0},{1},{2},{3}".format(x, y, w, h)

        if self.use_fullscreen():
            fullscreen_mode = self.options[Option.FULLSCREEN_MODE]
            if fullscreen_mode == "window":
                env["FSGS_FULLSCREEN_MODE"] = "1"
            elif fullscreen_mode == "fullscreen":
                env["FSGS_FULLSCREEN_MODE"] = "2"
            else:
                env["FSGS_FULLSCREEN_MODE"] = "3"

            for key in ["FSGS_GEOMETRY", "FSGS_WINDOW"]:
                if env.get(key, ""):
                    # Already specified externally, so we just use
                    # the existing values...
                    pass
                    print("using existing fullscreen window rect")
                else:
                    env[key] = "{0},{1},{2},{3}".format(x, y, w, h)
            print("fullscreen rect:", env.get("FSGS_GEOMETRY", ""))
            # SDL 1.2 multi-display support. Hopefully, SDL's display
            # enumeration is the same as QT's.
            env["SDL_VIDEO_FULLSCREEN_DISPLAY"] = str(self.screen_index())

            if fullscreen_mode == "window":
                print("using fullscreen window mode")
                # env["FSGS_FULLSCREEN"] = "window"
                env["FSGS_FULLSCREEN"] = "1"
            else:
                del env["FSGS_WINDOW"]
                if fullscreen_mode == "fullscreen":
                    print("using fullscreen (legacy) mode")
                    # env["FSGS_FULLSCREEN"] = "fullscreen"
                    env["FSGS_FULLSCREEN"] = "2"
                else:
                    print("using fullscreen desktop mode")
                    # env["FSGS_FULLSCREEN"] = "desktop"
                    env["FSGS_FULLSCREEN"] = "3"
        else:
            print("using window mode (no fullscreen)")
            env["FSGS_FULLSCREEN"] = "0"

        # if self.use_stretching():
        #     self.env["FSGS_STRETCH"] = "1"
        # else:
        #     self.env["FSGS_STRETCH"] = "0"

        if self.stretching() == self.STRETCH_FILL_SCREEN:
            self.emulator.env["FSGS_STRETCH"] = "1"
        elif self.stretching() == self.STRETCH_ASPECT:
            self.emulator.env["FSGS_STRETCH"] = "2"
        else:
            self.emulator.env["FSGS_STRETCH"] = "0"

        if self.border() == self.SMALL_BORDER:
            self.emulator.env["FSGS_BORDER"] = "1"
        elif self.border() == self.LARGE_BORDER:
            self.emulator.env["FSGS_BORDER"] = "2"
        else:
            self.emulator.env["FSGS_BORDER"] = "0"

        # env["FSGS_WINDOW_TITLE"] = ""
        self.set_environment_path(
            env, "FSGS_SCREENSHOTS_DIR", self.screenshots_dir()
        )
        self.set_environment_path(
            env, "FSGS_SCREENSHOTS_BASE", self.screenshots_base()
        )
        if self._model_name:
            env["FSGS_WINDOW_TITLE"] = self._model_name

        env.update(self.emulator.env)
        if not self.emulator.allow_home_access:
            env["HOME"] = self.home.path

        if self.options[Option.G_SYNC] == "0":
            print("[VIDEO] Disabling G-Sync (Might only work on Linux)")
            env["__GL_GSYNC_ALLOWED"] = "0"

        if not self._allow_gsync:
            # DOSBox-FS and Fuse-FS does not work nicely with G-SYNC yet.
            # Enabling G-SYNC causes stuttering.
            # Update: Should work fine with DOSBox-FS now...
            env["__GL_GSYNC_ALLOWED"] = "0"
            # Disable V-Sync
            env["__GL_SYNC_TO_VBLANK"] = "0"

        # Make sure we are allowed to flip buff ers faster than the screen
        # refresh rate, important for e.g. DOSBox @70Hz.
        # https://dri.freedesktop.org/wiki/ConfigurationOptions/
        # 1 = Application preference, default interval 0
        # Update: Not needed when emulators explicitly set swap interval 0.
        # env["vblank_mode"] = "1"

        self.update_environment_with_centering_info(env)

    def update_environment_with_centering_info(self, env):
        # FIXME: does not really belong here (dependency loop)
        from launcher.launcher_config import LauncherConfig
        from launcher.launcher_settings import LauncherSettings

        width = LauncherConfig.get("window_width") or LauncherSettings.get(
            "window_width"
        )
        height = LauncherConfig.get("window_height") or LauncherSettings.get(
            "window_height"
        )
        try:
            width = int(width)
        except:
            width = 960
        try:
            height = int(height)
        except:
            height = 540
        from launcher.ui.launcherwindow import LauncherWindow

        if LauncherWindow.current() is None:
            return

        main_w, main_h = LauncherWindow.current().get_size()
        main_x, main_y = LauncherWindow.current().get_position()

        x = main_x + (main_w - width) // 2
        y = main_y + (main_h - height) // 2

        # FIXME: REMOVE
        env["FSGS_WINDOW_X"] = str(x)
        env["FSGS_WINDOW_Y"] = str(y)

        # FIXME: REMOVE
        env["SDL_VIDEO_WINDOW_POS"] = str("{0},{1}".format(x, y))
        env["FSGS_WINDOW_POS"] = str("{0},{1}".format(x, y))

        env["FSGS_WINDOW_CENTER"] = "{0},{1}".format(
            main_x + main_w // 2, main_y + main_h // 2
        )

        # args.append("--window-x={0}".format(x))
        # args.append("--window-y={0}".format(y))
        # print("window position", env["SDL_VIDEO_WINDOW_POS"])
        # os.environ["SDL_VIDEO_WINDOW_POS"] = "{0},{1}".format(x, y)

    def emulator_skin_paths(self):
        left = self.temp_file("left.png").path
        left_overlay = self.temp_file("left-overlay.png").path
        right = self.temp_file("right.png").path
        right_overlay = self.temp_file("right-overlay.png").path
        return {
            "left": left,
            "left-overlay": left_overlay,
            "right": right,
            "right-overlay": right_overlay,
        }

    def prepare_emulator_skin(self, env=None, paths=None):
        if self._emulator_skin_prepared:
            return self._emulator_skin_prepared
        if paths is None:
            paths = self.emulator_skin_paths()
        with open(paths["left"], "wb") as f:
            f.write(Resources("fsgs").stream("res/emu/left.png").read())
        if env is not None:
            env["FSGS_BEZEL_LEFT"] = paths["left"]
            env["FSGS_SKIN_LEFT"] = paths["left"]
        if "left-overlay" in paths:
            with open(paths["left-overlay"], "wb") as f:
                f.write(
                    Resources("fsgs").stream("res/emu/left-overlay.png").read()
                )
            if env is not None:
                env["FSGS_BEZEL_LEFT_OVERLAY"] = paths["left-overlay"]
        with open(paths["right"], "wb") as f:
            f.write(Resources("fsgs").stream("res/emu/right.png").read())
        if env is not None:
            env["FSGS_BEZEL_RIGHT"] = paths["right"]
            env["FSGS_SKIN_RIGHT"] = paths["right"]
        if "right-overlay" in paths:
            with open(paths["right-overlay"], "wb") as f:
                f.write(
                    Resources("fsgs")
                    .stream("res/emu/right-overlay.png")
                    .read()
                )
            if env is not None:
                env["FSGS_BEZEL_RIGHT_OVERLAY"] = paths["right-overlay"]
        self._emulator_skin_prepared = paths
        return self._emulator_skin_prepared

    def start_emulator(
        self, emulator, args=None, env_vars=None, executable=None, cwd=None
    ):
        # if "/" in emulator:
        #     if not executable:
        #         executable = self.find_emulator_executable(emulator)
        #     if not executable:
        #         emulator = emulator.split("/")[-1]
        #
        # if not executable:
        #     executable = self.find_emulator_executable("fs-" + emulator)
        # if not executable:
        #     executable = self.find_emulator_executable(emulator)
        #
        # if not executable:
        #     raise Exception("could not find executable for " + emulator)

        args = []
        args.extend(self.emulator.args)

        if "SDL_VIDEODRIVER" in os.environ:
            print("SDL_VIDEODRIVER was present in environment, removing!")
            del os.environ["SDL_VIDEODRIVER"]

        env = os.environ.copy()
        FSUAE.add_environment_from_settings(env)
        self.update_environment(env)
        if self.bezel():
            self.prepare_emulator_skin(env)
        if env_vars:
            env.update(env_vars)
        print("")
        for key in sorted(env.keys()):
            print("[ENV]", key, ":", repr(env[key]))
        print("")
        for arg in args:
            print("[ARG]", repr(arg))
        print("")

        kwargs = {}
        if env is not None:
            kwargs["env"] = env
        if cwd is not None:
            kwargs["cwd"] = cwd
        else:
            kwargs["cwd"] = self.cwd.path
        print("[CWD]", kwargs["cwd"])
        print("")
        if System.windows:
            kwargs["close_fds"] = True
        # print(" ".join(args))
        current_task.set_progress(
            "Starting {emulator}".format(emulator=emulator)
        )
        # import subprocess
        # return subprocess.Popen(["strace", emulator.path] + args, **kwargs)
        process = emulator.popen(args, **kwargs)
        self.write_emulator_pid_file(self.emulator_pid_file(), process)
        return process

    def find_emulator_executable(self, name):
        # if os.path.isdir("../fs-uae/src"):
        #     # running from source directory, we then want to find locally
        #     # compiled binaries if available
        #     path = "../fs-uae/fs-uae"
        #     if System.windows:
        #         path += ".exe"
        #     if os.path.isfile(path):
        #         return path
        #     raise Exception("Could not find development FS-UAE executable")

        if "/" in name:
            package, name = name.split("/")
        else:
            package = name

        if System.windows:
            # first we check if the emulator is bundled inside the launcher
            # directory
            exe = os.path.join(fsboot.executable_dir(), package, name + ".exe")
            if not os.path.exists(exe):
                # for when the emulators are placed alongside the launcher /
                # game center directory
                exe = os.path.join(
                    fsboot.executable_dir(), "..", package, name + ".exe"
                )
            if not os.path.exists(exe):
                # when the emulators are placed alongside the fs-uae/ directory
                # containing launcher/, for FS-UAE Launcher & FS-UAE Arcade
                exe = os.path.join(
                    fsboot.executable_dir(), "..", "..", package, name + ".exe"
                )
            if not os.path.exists(exe):
                return None
            return exe
        elif System.macos:
            exe = os.path.join(
                fsboot.executable_dir(),
                "..",
                package + ".app",
                "Contents",
                "MacOS",
                name,
            )
            if not os.path.exists(exe):
                exe = os.path.join(
                    fsboot.executable_dir(),
                    "..",
                    "..",
                    "..",
                    package + ".app",
                    "Contents",
                    "MacOS",
                    name,
                )
            if not os.path.exists(exe):
                exe = os.path.join(
                    "/Applications",
                    package + ".app",
                    "Contents",
                    "MacOS",
                    name,
                )
            if not os.path.exists(exe):
                return None
            return exe

        if os.path.exists("/opt/{0}/bin/{1}".format(package, name)):
            return "/opt/{0}/bin/{1}".format(package, name)

        if package == name:
            for directory in os.environ["PATH"].split(":"):
                path = os.path.join(directory, name)
                if os.path.exists(path):
                    return path
        return None

    def get_game_refresh_rate(self):
        """Override in inherited classes to specify custom refresh rate."""
        if self.options[Option.REFRESH_RATE]:
            print("get_game_refresh_rate, refresh_rate was set")
            return float(self.options[Option.REFRESH_RATE])
        if self.options[Option.VIDEO_STANDARD] == "NTSC":
            return 59.94
        if self.options[Option.VIDEO_STANDARD] == "PAL":
            return 50
        variant_name = self.variant_name()
        print("get_game_refresh_rate, variant name =", variant_name)
        if "USA" in variant_name:
            return 59.94
        if "World" in variant_name:
            return 59.94
        if "Japan" in variant_name:
            return 59.94
        if "Europe" in variant_name:
            return 50
        print("get_game_refresh_rate: unknown")
        return 0

    def configure_vsync(self):
        print("\n" + "-" * 79 + "\n" + "CONFIGURE VSYNC")
        # print("^" * 80)
        allow_vsync = self.use_vsync()
        # if self.get_option('vsync'):
        if allow_vsync:
            # try:
            #     game_refresh = float(self.options["refresh_rate"])
            # except Exception:
            game_refresh = self.get_game_refresh_rate() or 0.0
            # if self.context.game.config.get('system', '') == 'NTSC':
            #     game_refresh = 60.0
            # else:
            #     game_refresh = 50.0

            # refresh_rate_tool = RefreshRateTool(
            #     game_platform=self.fsgc.game.platform.id,
            #     game_refresh=round(game_refresh))
            # refresh_rate_tool.set_best_mode()
            # if refresh_rate_tool.allow_vsync():
            #     print("setting vsync to true")
            #     self.set_vsync(True)
            #     return refresh_rate_tool.get_display_refresh()
            screen_refresh = self.screen_refresh_rate()
            if self.options[Option.ASSUME_REFRESH_RATE]:
                try:
                    screen_refresh = float(
                        self.options[Option.ASSUME_REFRESH_RATE]
                    )
                except ValueError:
                    print("Could not parse 'assume_refresh_rate' value")
                else:
                    print("Assuming screen refresh rate:", screen_refresh)

            print("Game refresh rate:", game_refresh)
            print("Screen refresh rate:", screen_refresh)
            if int(round(screen_refresh)) == int(round(game_refresh)):
                print("setting vsync to true")
                self.set_vsync(True)
                return True
        else:
            print("vsync is not enabled")
        print("setting vsync to false")
        self.set_vsync(False)
        return False

    def set_vsync(self, vsync):
        print("GameRunner - set vsync to", vsync)
        self.__vsync = vsync

    def abort(self):
        print("GameRunner.abort - WARNING: not implemented")

    def cheats_file(self, name):
        dev_path = os.path.join("..", "cheats", name)
        # FIXME: if development_mode ?
        if os.path.exists(dev_path):
            print("Found cheats file:", dev_path)
            return dev_path
        try:
            plugin = PluginManager.instance().plugin("Cheats")
        except LookupError:
            print("No cheats plugin found")
            return None
        path = plugin.data_file_path(name)
        if os.path.exists(path):
            print("Found cheats file:", path)
            return path
        print("No cheats file found:", name)
        return None

    @property
    def args(self):
        warnings.warn("deprecated", DeprecationWarning)
        return self.emulator.args

    @property
    def config(self):
        warnings.warn("deprecated", DeprecationWarning)
        return self.options

    @property
    def env(self):
        warnings.warn("deprecated", DeprecationWarning)
        return self.emulator.env

    @property
    def fsgs(self):
        warnings.warn("deprecated", DeprecationWarning)
        return self.fsgc

    def set_env(self, name, value):
        warnings.warn("deprecated", DeprecationWarning)
        self.emulator.env[name] = value

    def add_arg(self, *args):
        warnings.warn("deprecated", DeprecationWarning)
        self.emulator.args.extend(args)

    def create_temp_dir(self, suffix):
        warnings.warn("create_temp_dir is deprecated", DeprecationWarning)
        return self.temp_dir(suffix)

    def create_temp_file(self, suffix):
        warnings.warn("create_temp_file is deprecated", DeprecationWarning)
        return self.temp_file(suffix)


class Port(object):
    def __init__(self, name):
        self.number = 0
        self.name = name
        self.types = []
        self.index = 0
        self.device = None
        # Name of config option for device type
        self.type_option = ""
        # Name of config option for device
        self.device_option = ""

        # FIXME: remove
        self.device_id = None
        self.device_config = None

    @property
    def type(self):
        return self.types[self.index]["type"]

    @property
    def mapping_name(self):
        return self.types[self.index]["mapping_name"]

    @property
    def description(self):
        return self.types[self.index]["description"]


class TemporaryItem:
    def __init__(self, root, prefix="tmp", suffix="", directory=False):
        self.root = root
        self.prefix = prefix
        self.suffix = suffix
        self._path = None
        self.directory = directory

    def __del__(self):
        self.delete()

    def delete(self):
        print("[DRIVER] TemporaryItem.delete")
        assert self.root is None
        if self._path is None:
            return
        if os.environ.get("FSGS_CLEANUP", "") == "0":
            print("[DRIVER] NOTICE: keeping temp files around...")
            return
        if self._path:
            print("[DRIVER] Removing", self._path)
            shutil.rmtree(self._path)
            self._path = None

    @property
    def path(self):
        if self._path is None:
            if hasattr(self.root, "path"):
                # We want to delay temp dir creation as long as possible,
                # so we only call root.path when we have to.
                root = self.root.path
            else:
                root = self.root
            if self.directory:
                self._path = tempfile.mkdtemp(
                    prefix=self.prefix, suffix=self.suffix, dir=root
                )
            else:
                fd, self._path = tempfile.mkstemp(
                    prefix=self.prefix, suffix=self.suffix, dir=root
                )
                os.close(fd)
        return self._path


class TemporaryNamedItem:
    def __init__(self, root, name, directory=False):
        self.root = root
        self.name = name
        self._path = None
        self.directory = directory

    @property
    def path(self):
        if self._path is None:
            root = self.root.path
            self._path = os.path.join(root, self.name)
            if not os.path.exists(self._path):
                if self.directory:
                    os.makedirs(self._path)
                else:
                    with open(self.path, "wb"):
                        pass
        return self._path


class Emulator:
    def __init__(self, name):
        self.name = name
        self.args = []
        self.env = {}
        self.process = None
        self.allow_system_emulator = False
        self.allow_home_access = False


class GameFiles:
    def __init__(self, fsgc):
        self.fsgc = fsgc
        self._files = []

    def add(self, dest, source=None, sha1=None, description=None):
        self._files.append(
            {
                "source": source,
                "sha1": sha1,
                "dest": dest,
                "description": description,
            }
        )

    def install(self):
        for file in self._files:
            # uri = self.fsgc.file.find_by_sha1(tos_file.sha1)
            # if not uri:
            #     raise Exception(
            #             "Could not find {} (SHA-1: {})".format(
            #                 tos_file.name, tos_file.sha1))
            # self.files.add(self.tos_file.path, source=uri)
            if file["source"]:
                self.fsgc.file.copy_game_file(file["source"], file["dest"])
            else:
                assert file["sha1"]
                uri = self.fsgc.file.find_by_sha1(file["sha1"])
                if not uri:
                    description = file["description"] or os.path.basename(
                        file["dest"]
                    )
                    raise Exception(
                        "Could not find {} (SHA-1: {})".format(
                            description, file["sha1"]
                        )
                    )
                self.fsgc.file.copy_game_file(uri, file["dest"])



import atexit
import os
import shutil
import sys
import traceback
from configparser import ConfigParser, NoSectionError

# noinspection PyUnresolvedReferences
from typing import Dict, Tuple

import fsboot
from fsbc.signal import Signal
from fsbc.util import Version
from launcher.version import VERSION


class Settings(object):
    _instance = None

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = Settings()
        return cls._instance

    def __init__(self, app=None, path: str = None) -> None:
        self.app = app
        self.path = path
        self.values = {}  # type: Dict[str, str]
        self._provider = SettingsProvider()
        self._loaded = False
        self._loading = False
        self._atexit_registered = False
        self.verbose = True

    def set_path(self, path):
        self.path = path

    def set_provider(self, provider):
        self._provider = provider

    def get(self, key: str, default="") -> str:
        if not self._loading:
            self.load()
        return self.values.get(key, default)

    def set(self, key: str, value: str) -> None:
        if not self._loading:
            self.load()
        if self[key] == value:
            self.log_key_value(key, value, extra="(unchanged)")
            return
        if not self._loading and not self._atexit_registered:
            if self.verbose:
                print("[SETTINGS] Register atexit save path =", self.path)
            atexit.register(self.save)
            self._atexit_registered = True
        self.log_key_value(key, value)
        self.values[key] = value
        Signal("setting").notify(key, value)

    def __getitem__(self, key: str):
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> None:
        self.set(key, value)

    def load(self, force=False):
        # print("[SETTINGS] Load", self)
        if (not self._loaded) or force:
            try:
                self._loading = True
                self._provider.load(self, verbose=self.verbose)
            finally:
                self._loading = False
        self._loaded = True
        # print("[SETTINGS] Loaded, path is", self.path)

    def save(self, extra: Dict[str, str] = None) -> None:
        if self.verbose:
            print("[SETTINGS] Save", self)
        self._provider.save(self, extra)

    def log_key_value(self, key, value, extra=""):
        if (
            "username" in key
            or "password" in key
            or "auth" in key
            or "email" in key
        ):
            value = "*CENSORED*"
        if extra:
            extra = " " + extra
        if self.verbose:
            print("[SETTINGS] Set {} = {}{}".format(key, value, extra))


class SettingsProvider:
    """Provides default load and save behavior for Settings. You can subclass
    this and replace the provider on the Settings instance if you want
    customized loading/saving."""

    def load(self, settings, verbose=True):
        cp = ConfigParser(interpolation=None)
        cp.optionxform = str
        path = settings.path
        if settings.app and not path:
            path = settings.app.get_settings_path()
        if not path:
            if verbose:
                print("[SETTINGS] No settings path specified")
            path = os.path.join(fsboot.base_dir(), "Data", "Settings.ini")
            if verbose:
                print("[SETTINGS] Using default", path)
        if os.path.exists(path):
            if verbose:
                print("[SETTINGS] Loading from", path)
        else:
            if verbose:
                print("[SETTINGS] File", path, "does not exist")
        # Write current settings path back to Settings instance
        settings.path = path
        try:
            cp.read([path], encoding="UTF-8")
        except Exception as e:
            if verbose:
                print("[SETTINGS] Error loading", repr(e))
            return
        try:
            keys = cp.options("settings")
        except NoSectionError:
            return

        values = {}
        for key in sorted(keys):
            if key.startswith("__"):
                if verbose:
                    print("[SETTINGS] Ignoring", key)
                continue
            value = cp.get("settings", key)
            values[key] = value

        for arg in sys.argv:
            if arg.startswith("--settings:"):
                arg = arg[11:]
                key, value = arg.split("=", 1)
                key = key.replace("-", "_")
                values[key] = value

        for key in sorted(values.keys()):
            value = values[key]
            key = self.rewrite_key(key)
            settings.set(key, value)

        try:
            version = cp.get("launcher", "version")
        except NoSectionError:
            version = "0.0.0"
        try:
            fix_settings(settings, version)
        except Exception:
            print("[SETTINGS] Error fixing settings")
            traceback.print_exc()

    def rewrite_key(self, key):
        # FIXME: Should be moved to subclass
        try:
            return key_replacement_table[key]
        except KeyError:
            return key

    def save(self, settings, extra=None):
        partial_path = settings.path + ".partial"
        print("[SETTINGS] Writing to", partial_path)

        save_values = {}  # type: Dict[Tuple[str, str], str]
        for key, value in settings.values.items():
            if key.startswith("__"):
                # Temporary setting, do not save
                continue
            save_values[("settings", str(key))] = str(value)
        if extra is not None:
            for key, value in extra.items():
                try:
                    section, key = key.split("/")
                except ValueError:
                    section = "settings"
                save_values[(section, key)] = value

        cp = ConfigParser(interpolation=None)
        cp.optionxform = str
        # We want the settings section to be listed first.
        cp.add_section("settings")

        for (section, key) in sorted(save_values.keys()):
            value = save_values[(section, key)]
            if not value:
                # We do not need to write empty values, as non-existing
                # keys have an implicit empty value.
                continue
            if not cp.has_section(section):
                cp.add_section(section)
            cp.set(section, key, value)

        cp.add_section("launcher")
        cp.set("launcher", "version", VERSION)

        if not os.path.exists(os.path.dirname(partial_path)):
            os.makedirs(os.path.dirname(partial_path))
        with open(partial_path, "w", encoding="UTF-8", newline="\n") as f:
            cp.write(f)
        print("[SETTINGS] Moving to", settings.path)
        shutil.move(partial_path, settings.path)


def get(key: str) -> str:
    return Settings.instance().get(key)


def set(key: str, value: str) -> None:
    Settings.instance().set(key, value)


def load() -> None:
    Settings.instance().load()


def save() -> None:
    Settings.instance().save()


def unload() -> None:
    raise NotImplementedError("settings.unload")


def set_path(path: str) -> None:
    Settings.instance().set_path(path)


# FIXME: FS-UAE Launcher specific
def fix_settings(settings, version_str):
    """Fix settings, based on the version used top save the newly loaded
    settings."""
    old = Version(version_str)
    if old < Version("2.9.10"):
        if settings.get("fullscreen_mode"):
            print("[SETTINGS] Reverting fullscreen_mode to default")
            settings.set("fullscreen_mode", "")


# FIXME: FS-UAE Launcher specific
key_replacement_table = {
    "database_arcade": "arcade_database",
    "database_atari": "atari_database",
    "database_c64": "c64_database",
    "database_cpc": "cpc_database",
    "database_dos": "dos_database",
    "database_gb": "gb_database",
    "database_gba": "gba_database",
    "database_gbc": "gbc_database",
    "database_nes": "nes_database",
    "database_psx": "psx_database",
    "database_snes": "snes_database",
}



# from __future__ import division
# from __future__ import print_function
# from __future__ import absolute_import
# from __future__ import unicode_literals
#
# from weakref import ref
#
#
# class Signal(object):
#
#     def __init__(self):
#         self.listeners = []
#
#     def connect(self, listener):
#         print("connecting", listener)
#         self.listeners.append(ref(listener))
#
#     def notify(self, *args, **kwargs):
#         print("notify", self.listeners)
#         dead_refs = []
#         for listener_ref in self.listeners:
#             listener = listener_ref()
#             if listener is None:
#                 dead_refs.append(listener_ref)
#                 continue
#             print(listener, args, kwargs)
#             listener(*args, **kwargs)
#         for r in dead_refs:
#             self.listeners.remove(r)



import random

from arcade.glui.font import Font
from arcade.glui.opengl import gl, fs_emu_blending, fs_emu_texturing
from arcade.glui.render import Render
from arcade.glui.state import State
from arcade.glui.texture import Texture
from fsgs.platform import PlatformHandler

COORDS = 0.0, 1.0, 1.0, 0.0


def rectangle(x, y, w, h, z=0.0):
    # gl.glBegin(gl.GL_QUADS)
    gl.glVertex3f(x, y, z)
    gl.glVertex3f(x + w, y, z)
    gl.glVertex3f(x + w, y + h, z)
    gl.glVertex3f(x, y + h, z)
    # gl.glEnd()


def texture_rectangle(x, y, w, h, z=0.0, s1=0.0, s2=1.0, t1=1.0, t2=0.0):
    gl.glBegin(gl.GL_QUADS)
    gl.glTexCoord(s1, t1)
    gl.glVertex3f(x, y, z)
    gl.glTexCoord(s2, t1)
    gl.glVertex3f(x + w, y, z)
    gl.glTexCoord(s2, t2)
    gl.glVertex3f(x + w, y + h, z)
    gl.glTexCoord(s1, t2)
    gl.glVertex3f(x, y + h, z)
    gl.glEnd()


def static():
    s1 = random.uniform(0.0, 0.65)
    s2 = s1 + 0.35
    t1 = random.uniform(0.0, 0.65)
    t2 = t1 + 0.35
    # w, h = Texture.static.size
    # return Texture.static, (s1 * w, s2 * w, t1 * h, t2 * h)
    return Texture.static, (s1, s2, t1, t2)


last_item_id = None
item_time = 0


def render_bottom_bar_transparent():
    Render.get().hd_perspective()
    fs_emu_blending(True)
    Texture.bottom_bar.render(0, 0, 1920, 380, 0.3, opacity=1.0)
    item = State.get().current_menu.selected_item
    render_bottom_bar_text(item)


def render_bottom_bar():
    Render.get().hd_perspective()
    item = State.get().current_menu.selected_item
    render_bottom_bar_screens(item)


def render_bottom_bar_screens(item):
    global last_item_id
    global item_time
    if id(item) != last_item_id:
        last_item_id = id(item)
        item_time = State.get().time

    gl.glColor3f(1.0, 1.0, 1.0)
    fs_emu_texturing(True)
    fs_emu_blending(False)

    for i in range(3, -1, -1):
        if State.get().time - item_time > 0.1 + 0.05 * i:
            texture = item.get_screen_texture(2 + i)
        else:
            texture = None
        if texture:
            # coords = (0, texture.w, texture.h, 0)
            coords = COORDS
        else:
            texture, coords = static()
            # While rendering static, we must continue rendering.
            Render.get().dirty = True

        texture.bind()
        texture_rectangle(4 + 539 + 345 * i, 34, 308, 193, 0.2, *coords)
    if State.get().time - item_time > 0.1:
        texture = item.get_screen_texture(1)
    else:
        texture = None
    if texture:
        coords = COORDS
    else:
        texture, coords = static()
        # While rendering static, we must continue rendering.
        Render.get().dirty = True
    texture.bind()

    texture_rectangle(34, 34, 472, 295, 0.2, *coords)


def render_bottom_bar_text(item):
    strength = 0.9
    x = 544
    y = 290

    title = item.title.upper()
    title = title.strip()
    if "[" in title:
        title, subtitle = title.rsplit("[", 1)
        title = title.strip()
        subtitle = subtitle.strip()
        if subtitle.endswith("]"):
            subtitle = subtitle[:-1]
    else:
        subtitle = ""
    if title:
        Render.get().text(
            title,
            Font.title_font,
            x,
            y,
            1920 - x - 170,
            shadow=True,
            color=(1.0, 1.0, 1.0, 1.0 * strength),
        )
        if subtitle:
            color = (0x6E / 0xFF, 0x8B / 0xFF, 0x96 / 0xFF, 0.75)
            tw, th = Render.get().measure_text(title, Font.title_font)
            Render.get().text(
                subtitle,
                Font.title_font,
                x + tw + 20,
                y,
                1920 - x - 170,
                shadow=True,
                color=color,
            )

    color = (0xC4 / 0xFF, 0xD7 / 0xFF, 0xDE / 0xFF, 1.0)
    year_text = str(
        getattr(State.get().current_menu.selected_item, "year", "") or ""
    ).upper()
    if year_text:
        tw, th = Render.get().measure_text(year_text, Font.title_font)
        Render.get().text(
            year_text,
            Font.title_font,
            1920 - 30 - tw,
            y,
            0,
            shadow=True,
            color=color,
        )
    color = (0x6E / 0xFF, 0x8B / 0xFF, 0x96 / 0xFF, 1.0)
    y = 258
    text_str = ""
    companies = set()
    publisher_text = (
        getattr(State.get().current_menu.selected_item, "publisher", "") or ""
    ).upper()
    for text in publisher_text.split("/"):
        text = text.strip()
        if text:
            if text not in companies:
                text_str = text_str + u" \u00b7 " + text
                companies.add(text)
    developer_text = (
        getattr(State.get().current_menu.selected_item, "developer", "") or ""
    ).upper()
    for text in developer_text.split("/"):
        text = text.strip()
        if text:
            if text not in companies:
                text_str = text_str + u" \u00b7 " + text
                companies.add(text)
    if len(text_str) > 3:
        text_str = text_str[3:]  # remove initial middle dot
        tw, th = Render.get().measure_text(text_str, Font.subtitle_font)
        Render.get().text(
            text_str,
            Font.subtitle_font,
            1920 - 30 - tw,
            y,
            0,
            shadow=True,
            color=color,
        )

    platform_str = str(
        getattr(State.get().current_menu.selected_item, "platform", "") or ""
    )
    if platform_str:
        platform_str = PlatformHandler.get_platform_name(platform_str).upper()
    if len(platform_str) >= 3:
        # tw, th = Render.get().measure_text(platform_str, Font.subtitle_font)
        Render.get().text(
            platform_str, Font.subtitle_font, x, y, 0, shadow=True, color=color
        )
    return



import time
import traceback

from arcade.resources import logger
from fsbc.util import memoize
from fsgs.input.eventlistener import EventListener
from fsgs.input.inputdevice import InputDevice
from fsgs.input.keyboard import Keyboard
from fsgs.input.devicemanager import DeviceManager

REPEAT_THRESHOLD = 0.300
REPEAT_INTERVAL = 0.075


def get_controller_config(name, sdl_name, axes=0, hats=0, buttons=0, balls=0):
    try:
        # device id must end with #something (really a device number,
        # but can be anything
        device = InputDevice(
            "menu",
            name + " #MENU",
            [],
            sdl_name=sdl_name,
            version=2,
            axes=axes,
            hats=hats,
            buttons=buttons,
            balls=balls,
        )
        config = device.get_config()
    except Exception:
        print("error initializing device " + repr(name) + " for menu")
        traceback.print_exc()
        # print(repr(e))
        return None
    # config_inv = []
    for key, val in list(config.items()):
        val = val.upper()
        config[key] = val
        config[val] = key
    return config


class InputHandler(object):
    last_device = ""
    joystick_hash = None
    last_joystick_check = 0
    current_button = None
    repeat_info = None
    repeatable_buttons = ["UP", "DOWN", "LEFT", "RIGHT", "BACK"]
    # last_joystick_count = 0
    joysticks = []
    axis_status = {}
    first_init = True
    device_sdl_names = {}
    key_table = {}
    event_listener = None
    event_queue = []
    text_event_queue = []
    inhibit = False

    @classmethod
    def set_inhibited(cls, inhibit=True):
        cls.inhibit = inhibit

    @classmethod
    def get_virtual_button(cls, event):
        # print("get_virtual_button", event)
        assert isinstance(event, dict)
        if event["type"] in ["key-down", "key-up"]:
            try:
                key = Keyboard.key(event)
            except Exception:
                traceback.print_exc()
                # print(repr(e))
                return None, None
            if key.name == "SDLK_NO_KEY":
                return None, None
            try:
                button, device_id = cls.key_table[key.name]
            except Exception:
                traceback.print_exc()
                return None, None

            # if button:
            #     #cls.last_device = "KEYBOARD #1"
            #     cls.last_device = device_id
            # print("get_virtual_button", button, cls.last_device)
            return button, device_id
        elif event["type"] in [
            "joy-button-down",
            "joy-button-up",
            "joy-axis-motion",
            "joy-hat-motion",
        ]:
            if event["type"] == "joy-axis-motion":
                # print(event)
                # if event.value > -0.5 and event.value < 0.5:
                #     return None
                # sign = "-" if event.value < 0 else "+"
                sign = "neg" if event["state"] < 0 else "pos"
                cfg_name = "axis_{0}_{1}".format(event["axis"], sign)
            elif event["type"] == "joy-hat-motion":
                print(event)
                if event["state"] & 8:
                    value = "left"
                elif event["state"] & 2:
                    value = "right"
                elif event["state"] & 4:
                    value = "down"
                elif event["state"] & 1:
                    value = "up"
                else:
                    value = "0"
                cfg_name = "hat_{0}_{1}".format(event["hat"], value)
            else:
                cfg_name = "button_{0}".format(event["button"])

            # try:
            #     joystick = cls.joysticks[event["device"]]
            # except IndexError:
            #     return None, None
            # print(joystick)
            # controller = get_controller_instance(joystick.name)
            # if controller is None:
            #     return None
            # #klass.get_
            # config = controller.get_config()
            # config_inv = controller.get_config_inverted()

            config = cls.get_controller_config(event["device"])
            if config is None:
                return None, None

            joystick = cls.joysticks[event["device"]]
            try:
                button = config[cfg_name]
            except KeyError:
                return None, None

            # if button == "START":
            #     combine = "SELECT"
            # elif button == "SELECT":
            #     combine = "START"
            # else:
            #     combine = None
            # if combine:
            #     try:
            #         menu = False
            #         #back = False
            #         skip_left = False
            #         skip_right = False
            #         b = config[combine]
            #         if b.startswith("button_"):
            #             b = int(b[7:])
            #             if joystick.joy_object.get_button(b):
            #                 menu = True
            #         b = config["SKIP_LEFT"]
            #         if b.startswith("button_"):
            #             b = int(b[7:])
            #             if joystick.joy_object.get_button(b):
            #                 skip_left = True
            #         b = config["SKIP_RIGHT"]
            #         if b.startswith("button_"):
            #             b = int(b[7:])
            #             if joystick.joy_object.get_button(b):
            #                 skip_right = True
            #         if menu:
            #             if skip_left and skip_right:
            #                 print("ABORT")
            #                 return "ABORT", joystick.id
            #             print("MENU")
            #             return "MENU", joystick.id
            #     except KeyError:
            #         pass

            # if button:
            #     print("setting last device to", joystick.id, "event", event)
            #     cls.last_device = joystick.id

            return button, joystick["id"]

    @classmethod
    @memoize
    def get_controller_config(cls, index):
        try:
            joystick = cls.joysticks[index]
        except IndexError:
            print("no joystick at index", index)
            return None

        return get_controller_config(
            joystick["name"],
            joystick["name"],
            axes=joystick["axes"],
            hats=joystick["hats"],
            buttons=joystick["buttons"],
            balls=joystick["balls"],
        )

    @classmethod
    def handle_device_event(cls, event):
        if event["type"] == "joy-device-added":
            print("ADDED device", event)
            cls.joysticks.append(event)

    @classmethod
    def handle_event(cls, event):
        assert isinstance(event, dict)
        down_event = False
        up_event = False
        if event["type"] == "key-down" or event["type"] == "joy-button-down":
            virtual_button, device_id = cls.get_virtual_button(event)
            down_event = True
        elif event["type"] == "key-up" or event["type"] == "joy-button-up":
            virtual_button, device_id = cls.get_virtual_button(event)
            up_event = True
        elif event["type"] == "joy-hat-motion":
            axis_name = "%d_%d" % (event["device"], 1000 + event["hat"])
            try:
                if cls.axis_status[axis_name] != event["state"]:
                    up_event = True
                    cls.repeat_info = None
            except KeyError:
                pass
            cls.axis_status[axis_name] = event["state"]
            virtual_button, device_id = cls.get_virtual_button(event)
            if virtual_button:
                down_event = True
        elif event["type"] == "joy-axis-motion":
            virtual_button, device_id = cls.get_virtual_button(event)
            axis_name = "%d_%d" % (event["device"], event["axis"])
            # if event.axis == 6:
            #     print(axis_name, virtual_button, event.value)
            if event["state"] < -0.66 * 32768:
                if cls.axis_status.setdefault(axis_name, 0) != -1:
                    cls.axis_status[axis_name] = -1
                    down_event = True
            elif event["state"] > 0.66 * 32768:
                if cls.axis_status.setdefault(axis_name, 0) != 1:
                    cls.axis_status[axis_name] = 1
                    down_event = True
            elif -0.33 * 32768 < event["state"] < 0.33 * 32768:
                if cls.axis_status.setdefault(axis_name, 0) != 0:
                    # print("axis status", cls.axis_status[axis_name])
                    cls.axis_status[axis_name] = 0
                    up_event = True
        elif event["type"] == "joy-device-added":
            return cls.handle_device_event(event)
        else:
            return
        if down_event:
            if virtual_button:
                # print(virtual_button, down_event, up_event)
                cls.set_current_button(virtual_button)
                if virtual_button in cls.repeatable_buttons:
                    t = time.time()
                    cls.repeat_info = [virtual_button, t, t]
                # cls.last_device_id = device_id
                cls.last_device = device_id
                # print("last device is", device_id)
        elif up_event:
            if virtual_button:
                # print("up event", event)
                cls.repeat_info = None

    @classmethod
    def update(cls):
        for event in cls.pop_all_events():
            cls.handle_event(event)
        cls.update_joysticks()
        cls.update_repeat()

    @classmethod
    def update_repeat(cls):
        t = time.time()
        if cls.repeat_info:
            if t - cls.repeat_info[1] > REPEAT_THRESHOLD:
                if t - cls.repeat_info[2] > REPEAT_INTERVAL:
                    cls.set_current_button(cls.repeat_info[0])
                    cls.repeat_info[2] = t

    @classmethod
    def update_joysticks(cls):
        if cls.event_listener is None:
            return
        while True:
            event = cls.event_listener.get_next_event()
            if not event:
                break
            InputHandler.handle_event(event)

    @classmethod
    def add_event(cls, event):
        if event["type"] == "text":
            cls.text_event_queue.append(event)
        elif event["type"] in ["mouse-motion", "mouse-press", "mouse-release"]:
            # text events basically now means events handled by the main
            # gui loop
            cls.text_event_queue.append(event)
        else:
            cls.event_queue.append(event)

    @classmethod
    def pop_event(cls):
        return cls.event_queue.pop()

    @classmethod
    def pop_all_events(cls):
        events = cls.event_queue[:]
        cls.event_queue.clear()
        return events

    @classmethod
    def pop_all_text_events(cls):
        events = cls.text_event_queue[:]
        cls.text_event_queue.clear()
        return events

    @classmethod
    def set_current_button(cls, button):
        if cls.inhibit:
            print("[INPUT] Inhibited: Ignoring", button)
            return
        cls.current_button = button

    @classmethod
    def peek_button(cls):
        button = cls.current_button
        return button

    @classmethod
    def get_button(cls):
        button = cls.current_button
        cls.current_button = None
        return button

    @classmethod
    def clear_current_button(cls):
        cls.current_button = None
        cls.repeat_info = None

    @classmethod
    def close(cls):
        cls._close_joysticks()

    @classmethod
    def open(cls):
        cls._open_joysticks()
        if cls.event_listener is None:
            cls.event_listener = EventListener()

    @classmethod
    def get_device_manager(cls):
        return cls.event_listener.device_manager

    @classmethod
    def reinit_joysticks(cls):
        print("\n" + "-" * 79 + "\n" + "INPUTHANDLER REINIT JOYSTICKS")
        print("reinit_joysticks")
        cls.axis_status = {}

        # for joystick in cls.joysticks:
        #     print("    un-initialize", joystick)
        #     joystick.joy_object.quit()

        # FIXME: stop using the Device Manager instance here?

        device_manager = DeviceManager.instance()
        devices = device_manager.get_devices()

        cls.key_table = {}
        for device in devices:
            print(device, "is keyboard device?", device.is_keyboard())
            if not device.is_keyboard():
                continue
            config = get_controller_config(
                device.name,
                device.sdl_name,
                axes=device.axes,
                hats=device.hats,
                buttons=device.buttons,
                balls=device.balls,
            )
            print(config)
            if not config:
                continue
            for key, value in config.items():
                if key.startswith("key_"):
                    cls.key_table["SDLK_" + key[4:].upper()] = (
                        value.upper(),
                        device.id,
                    )
        print("[INPUT] Key table:", cls.key_table)

    @classmethod
    def _open_joysticks(cls):
        logger.debug("InputHandler._open_joysticks")
        cls.reinit_joysticks()

    @classmethod
    def _close_joysticks(cls):
        pass



import logging
import os
import sys
import threading
import time

import fsboot
from fsbc.paths import Paths
from fsbc.settings import Settings
from fsbc.system import windows, macosx
from fsbc.util import memoize

logger = logging.getLogger("APP")


class Application(object):
    app_name = ""
    app_version = ""
    _instance = None

    @classmethod
    def instance(cls):
        """
        :rtype : Application
        """
        return cls._instance

    @classmethod
    def get_instance(cls):
        """
        :deprecated
        """
        return cls._instance

    @classmethod
    def get(cls):
        """
        :deprecated
        """
        return cls._instance

    @classmethod
    def set_instance(cls, instance):
        """
        :type instance: Application
        """
        if cls._instance:
            raise RuntimeError("An application instance already exists")
            # logger.warning("An application instance already exists")

        # noinspection PyAttributeOutsideInit
        cls._instance = instance
        # Settings.instance().set_path(instance.get_settings_path())

    def __init__(self, name="", version=""):
        # if Application.instance is not None:
        #     raise Exception("An application instance already exists")
        # Application.instance = self
        self.stop_flag = False
        self.name = name or Application.app_name
        self.version = version or Application.app_version
        self.__settings = None
        self._data_dirs = None
        Application.set_instance(self)

    def __del__(self):
        self.destroy()

    def destroy(self):
        self.stop()
        if self == Application._instance:
            Application._instance = None

    def set_stop_flag(self):
        self.stop_flag = True

    def stop(self):
        self.set_stop_flag()

    def wait(self):
        self.wait_for_threads()

    @classmethod
    def wait_for_threads(cls):
        logger.debug("Application.wait_for_threads")

        def get_threads():
            result = []
            for thread in threading.enumerate():
                if thread.daemon:
                    continue
                if not thread.is_alive():
                    continue
                result.append(repr(thread))
            return result

        t_list = get_threads()
        logger.debug(repr(t_list))
        while len(t_list) > 1:
            t_list_2 = get_threads()
            if t_list_2 != t_list:
                t_list = t_list_2
                logger.debug(repr(t_list))
            time.sleep(0.1)

    def stopping(self):
        return self.stop_flag

    def run_in_main(self, function, *args, **kwargs):
        raise NotImplementedError("Application.run_in_main")

    # def timer(self, timeout, function, *args, **kwargs):
    #     raise NotImplementedError("Application.call_later")

    @staticmethod
    @memoize
    def executable_dir():
        return fsboot.executable_dir()

    def cache_dir(self):
        # FIXME:
        return os.path.join(Paths.get_base_dir(), "Cache")

    @memoize
    def data_dirs(self):
        if self._data_dirs is not None:
            return self._data_dirs
        data_dirs = []
        base_dirs = []

        if len(sys.argv) > 0:
            script_dir = os.path.dirname(sys.argv[0])
            base_dirs.append(os.path.join(script_dir, "share"))

        data_dirs.append(self.executable_dir())
        if windows:
            base_dirs.append(os.path.join(self.executable_dir(), "share"))
        elif macosx:
            base_dirs.append(
                os.path.join(self.executable_dir(), "..", "Resources", "share")
            )
        else:
            # FIXME: $XDG_DATA_DIRS, $XDG_DATA_HOME
            base_dirs.append(
                os.path.normpath(
                    os.path.join(self.executable_dir(), "..", "share")
                )
            )
        for dir_name in base_dirs:
            data_dir = os.path.join(dir_name, self.name)
            logger.debug("* checking for data dir %s", data_dir)
            if os.path.exists(data_dir):
                data_dirs.append(data_dir)

        if macosx:
            data_dir = os.path.join(self.executable_dir(), "..", "Resources", "Data")
            if os.path.exists(data_dir):
                data_dirs.append(data_dir)

        self._data_dirs = data_dirs
        logger.debug("data dirs: %s", repr(data_dirs))
        return data_dirs

    @memoize
    def data_file(self, name):
        """Looks up an application data file in several places depending on
        platform.
        """
        for data_dir in self.data_dirs():
            path = os.path.join(data_dir, name)
            # print("- checking", path)
            if os.path.exists(path):
                return path
        raise LookupError(name)

    def get_settings_path(self):
        logger.warning("Application.get_settings_path not implemented")

    @property
    def settings(self):
        return Settings.instance()


def call_after(func, *args, **kwargs):
    return Application.instance().run_in_main(func, *args, **kwargs)


class ApplicationInstanceProxy(object):
    def __getattr__(self, name):
        return getattr(Application.instance(), name)

    def __setattr__(self, name, value):
        raise Exception("not currently allowed to set attributes on app")
        # return setattr(Application.instance(), name, value)


app = ApplicationInstanceProxy()



from .contextaware import ContextAware


class BaseContext(ContextAware):
    def __init__(self, main_context):
        ContextAware.__init__(self, main_context)

    @property
    def config(self):
        return self.context.config

    def set_config(self, *args):
        return self.context.config.set(*args)



import fsui
from launcher.i18n import gettext
from launcher.settings.settings_page import SettingsPage


class MouseSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("mouse-settings", "pkg:workspace")
        gettext("Mouse Settings")
        title = gettext("Mouse")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option("automatic_input_grab")
        self.add_option("initial_input_grab")
        self.add_option("middle_click_ungrab")
        self.add_option("mouse_speed")
        self.add_option("joystick_port_0_autoswitch")



from arcade.glui.opengl import gl


class DisplayLists(object):
    display_lists = {}

    @classmethod
    def clear(cls):
        for listid in cls.display_lists.values():
            gl.glDeleteLists(listid, 1)
        cls.display_lists.clear()

    @classmethod
    def call_or_create(cls, *params):
        try:
            gl.glCallList(cls.display_lists[params])
            return False
        except KeyError:
            display_list = gl.glGenLists(1)
            print("creating display list for params", params)
            gl.glNewList(display_list, gl.GL_COMPILE_AND_EXECUTE)
            cls.display_lists[params] = display_list
            return True



import weakref
from fsui.qt import Qt, QPoint, QCursor
from fsui.qt import QFontMetrics, QPalette, QWidget
from ..common.layout import Layout
from .DrawingContext import Font
from .Color import Color


class MixinBase(object):
    def __init__(self):
        # self._timer_started = False
        pass

    def init_mixin_base(self):
        pass

    #     self._timer_started = False
    # def set_timer(self, interval):
    #     assert not self._timer_started
    #     # noinspection PyUnresolvedReferences
    #     self.startTimer(interval)
    #     self._timer_started = True
    #
    # def on_timer(self):
    #     pass
    #
    # # noinspection PyPep8Naming
    # def timerEvent(self, _):
    #     # import traceback
    #     # traceback.print_stack()
    #     print(self, _)
    #     self.on_timer()


# noinspection PyPep8Naming
class WidgetMixin(MixinBase):
    def __init__(self, *_):
        # MixinBase.__init__(self)
        self._parent = None
        self._window = None

    def init_widget(self, parent):
        self.init_mixin_base()
        # self.layout = None
        self._parent = weakref.ref(parent)
        # noinspection PyProtectedMember
        self._window = parent._window
        # widget = getattr(self, "_widget", self)
        # widget.move(10000, 10000)

        widget = getattr(self, "_widget", self)
        # if self.get_window() != widget:
        assert isinstance(widget, QWidget)
        widget.installEventFilter(self.get_window())

        widget.destroyed.connect(self.on_destroy)

    def on_destroy(self):
        print("on_destroy")

    def eventFilter(self, obj, event):
        return False

    def get_parent(self):
        if self._parent is None:
            return None
        # noinspection PyCallingNonCallable
        return self._parent()

    def get_container(self):
        widget = getattr(self, "_widget", self)
        return widget

    def get_font(self):
        widget = getattr(self, "_widget", self)
        return Font(widget.font())

    def set_font(self, font):
        widget = getattr(self, "_widget", self)
        widget.setFont(font.font)

    def is_visible(self):
        widget = getattr(self, "_widget", self)
        return widget.isVisible()

    def visible(self):
        return self.is_visible()

    def measure_text(self, text):
        widget = getattr(self, "_widget", self)
        font = widget.font()
        metrics = QFontMetrics(font)
        return metrics.width(text), metrics.height()

    def set_hand_cursor(self):
        widget = getattr(self, "_widget", self)
        widget.setCursor(Qt.PointingHandCursor)

    def set_normal_cursor(self):
        widget = getattr(self, "_widget", self)
        widget.setCursor(Qt.ArrowCursor)

    def is_enabled(self):
        widget = getattr(self, "_widget", self)
        return widget.isEnabled()

    def set_visible(self, visible):
        return self.show_or_hide(visible)

    def show_or_hide(self, show):
        widget = getattr(self, "_widget", self)
        if show:
            widget.show()
        else:
            widget.hide()

    # def show(self, show=True):
    #     widget = getattr(self, "_widget", None)
    #     if widget is None:
    #         if show:
    #             widget.show()
    #         else:
    #             widget.hide()
    #     else:
    #         if show:
    #             super().show()
    #         else:
    #             super().hide()

    def focus(self):
        widget = getattr(self, "_widget", self)
        widget.setFocus()

    # def get_parent(self):
    #     # widget = getattr(self, "_widget", self)
    #     return self.parent()

    def refresh(self):
        widget = getattr(self, "_widget", self)
        return widget.update()

    def set_min_width(self, width):
        # noinspection PyAttributeOutsideInit
        self.min_width = width

    def set_min_height(self, height):
        # noinspection PyAttributeOutsideInit
        self.min_height = height

    def get_min_width(self):
        widget = getattr(self, "_widget", self)
        width = 0
        if hasattr(self, "min_width"):
            if self.min_width:
                width = max(self.min_width, width)
        if hasattr(self, "layout") and isinstance(self.layout, Layout):
            width = max(self.layout.get_min_width(), width)
            return width
        # result = max(width, widget.minimumSizeHint().width())
        # if widget.maximumWidth():
        #     print(widget.maximumWidth())
        #     return min(result, widget.maximumWidth())
        # return min(result, widget.maximumWidth())
        # return result
        result = max(width, widget.minimumSizeHint().width())
        return min(result, widget.maximumWidth())
        # return max(width, widget.minimumWidth())

    def get_min_height(self):
        widget = getattr(self, "_widget", self)
        assert isinstance(widget, QWidget)
        height = 0
        if hasattr(self, "min_height"):
            if self.min_height:
                height = max(self.min_height, height)
        if hasattr(self, "layout") and isinstance(self.layout, Layout):
            height = max(self.layout.get_min_height(), height)
            return height
        return max(height, widget.minimumSizeHint().height())
        # return max(height, widget.minimumHeight())

    def set_position(self, position, y=None):
        widget = getattr(self, "_widget", self)
        if y is None:
            widget.move(*position)
        else:
            widget.move(position, y)

    def set_size(self, size):
        widget = getattr(self, "_widget", self)
        widget.resize(*size)

    def set_position_and_size(self, position, size):
        widget = getattr(self, "_widget", self)
        widget.move(*position)
        widget.resize(*size)

    def get_window(self):
        # noinspection PyCallingNonCallable
        return self._window()
        # # widget = getattr(self, "_widget", self)
        # parent = self
        # # while parent.parentWidget() and not isinstance(parent, QMainWindow) \
        # #         and not isinstance(parent, QDialog):
        # #     parent = parent.parentWidget()
        # print(parent)
        # while parent.get_parent():
        #     parent = parent.get_parent()
        # return parent
        # #while self.parent:
        # #    parent = self.parent
        # #return parent

    def get_position(self):
        widget = getattr(self, "_widget", self)
        return widget.x(), widget.y()

    def get_size(self):
        widget = getattr(self, "_widget", self)
        return widget.width(), widget.height()

    def set_tool_tip(self, tool_tip):
        widget = getattr(self, "_widget", self)
        widget.setToolTip(tool_tip)

    def set_tooltip(self, tool_tip):
        self.set_tool_tip(tool_tip)

    def set_enabled(self, enable=True):
        widget = getattr(self, "_widget", self)
        widget.setEnabled(enable)

    # DEPRECATED
    def disable(self):
        return self.enable(False)

    # DEPRECATED
    def enable(self, enable=True):
        self.set_enabled(enable)

    def on_resize(self):
        if hasattr(self, "layout") and isinstance(self.layout, Layout):
            self.layout.set_size(self.get_size())
            self.layout.update()

    def get_background_color(self):
        # noinspection PyUnresolvedReferences
        return Color(self.palette().color(QPalette.Window))

    def set_background_color(self, color):
        # noinspection PyUnresolvedReferences
        self.setAutoFillBackground(True)
        widget = getattr(self, "_widget", self)
        p = widget.palette()
        # noinspection PyUnresolvedReferences
        p.setColor(self.backgroundRole(), color)
        # p.setColor(self.backgroundRole(), QColor(0xff, 0xaa, 0xaa))
        widget.setPalette(p)

    # def on_left_up(self):
    #     pass

    def popup_menu(self, menu, pos=(0, 0), blocking=True):
        # popup does not block, and if menu goes out of the scope of the
        # caller, it will disappear (unless we keep a reference here
        # FIXME: using exec now
        # self.__last_popup_menu = menu
        widget = getattr(self, "_widget", self)
        global_pos = widget.mapToGlobal(QPoint(pos[0], pos[1]))
        menu.set_parent(self)
        if blocking:
            menu.qmenu.exec(global_pos)
        else:
            menu.qmenu.popup(global_pos)

        # Firing off a fake mouse left up event really assumes that the
        # menu was opened with a left down event, so implementation isn't
        # ideal. Better to add a listener to menu close instead.

        # if hasattr(self, "on_left_up"):
        #     self.on_left_up()

    def is_mouse_over(self):
        # noinspection PyArgumentList
        c = QCursor.pos()
        # noinspection PyUnresolvedReferences
        p = self.mapToGlobal(QPoint(0, 0))
        s = self.get_size()
        if p.x() <= c.x() < p.x() + s[0] and p.y() <= c.y() < p.y() + s[1]:
            return True
        return False
